/*
 * Virtual to Real Hardware Interface.
 * (c) 2024 by Jerzy Kut <mono@protolabo.pl>
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <termios.h>
#include <fcntl.h>
#include <string.h>

#include "vrhi.h"


/*#define DEBUG*/


typedef enum bool
{
	FALSE = (0 == 1),
	TRUE = (1 == 1)
}
bool_t;

/* Maximum speed of communication. */
#define MAXSPEED B921600	/*B4000000*/

/* Timeout for receiving data. */
#define INTIMEOUT 3.0*10

/* No file descriptor assigned. */
#define FDNONE -1

/* Version of API. */
const int vrhi_version = 1;
const int vrhi_revision = 3;

/* Device signature. */
static const char SIGNATURE[] = { 'V', 'R', 'H', 'I' };
/* Device platform. */
static const char PLATFORM[] = { 'A', '8' };


/* File descriptor for the serial device /dev/tty*. */
static int file_descriptor = FDNONE;

/* Communication buffer. */
static char communication_buffer[1+BANK_SIZE+1];


#ifdef DEBUG
static void printf_buffer(const char *prefix, int count)
{
	int i;

	printf(prefix);
	if (count > 0)
	{
		for (i = 0; i < count; i++)
		{
			printf(" $%02x", (unsigned char)communication_buffer[i]);
		}
	}
	else
	{
		printf(" no data");
	}
	printf("\n");
}
#endif

static int block_read(void *buffer, ssize_t size)
{
	ssize_t result, count;

	count = 0;
	while (count < size)
	{
		result = read(file_descriptor, buffer, size - count);
		if (result <= 0)
		{
			return result;
		}
		buffer += result;
		count += result;
	}
	return count;
}

static int block_write(void *buffer, ssize_t size)
{
	ssize_t result, count;

	count = 0;
	while (count < size)
	{
		result = write(file_descriptor, buffer, size - count);
		if (result <= 0)
		{
			return result;
		}
		buffer += result;
		count += result;
	}
	return count;
}

#ifdef DEBUG
static char *tospeedstring(speed_t speed)
{
	switch (speed)
	{
	/* c_cflag bit meaning */
	case B0:	return "hang up";

	case B50:	return "50";
	case B75:	return "75";
	case B110:	return "110";
	case B134:	return "134";
	case B150:	return "150";
	case B200:	return "200";
	case B300:	return "300";
	case B600:	return "600";
	case B1200:	return "1200";
	case B1800:	return "1800";
	case B2400:	return "2400";
	case B4800:	return "4800";
	case B9600:	return "9600";
	case B19200:	return "19200";
	case B38400:	return "38400";
	/* extra output baud rates (not in POSIX) */
	case B57600:	return "57600";
	case B115200:	return "115200";
	case B230400:	return "230400";
	case B460800:	return "460800";
	case B500000:	return "500000";
	case B576000:	return "576000";
	case B921600:	return "921600";
	case B1000000:	return "1000000";
	case B1152000:	return "1152000";
	case B1500000:	return "1500000";
	case B2000000:	return "2000000";
	case B2500000:	return "2500000";
	case B3000000:	return "3000000";
	case B3500000:	return "3500000";
	case B4000000:	return "4000000";

	default:	return "unknown";
	}
}
#endif

vrhi_status_t vrhi_open( char *serial )
{
	if (file_descriptor < 0)
	{
		struct termios terminal;
#ifdef DEBUG
		speed_t ispeed, ospeed;
#endif
		int status;

		file_descriptor = open(serial, O_RDWR | O_NOCTTY);
		if (file_descriptor < 0)
		{
			return EERRNO;
		}

		/* 1 start bit, 8 data bits, 1 stop bit, no parity, 921600 baud, 3s timeout */
		status = tcgetattr(file_descriptor, &terminal);
		if (status < 0)
		{
			return EERRNO;
		}
		cfmakeraw(&terminal);
		terminal.c_cc[VMIN] = 0;		/* min number of bytes */
		terminal.c_cc[VTIME] = INTIMEOUT;	/* timeout in dsecs */
		status = cfsetispeed(&terminal, MAXSPEED);
		if (status < 0)
		{
			return EERRNO;
		}
		status = cfsetospeed(&terminal, MAXSPEED);
		if (status < 0)
		{
			return EERRNO;
		}
		status = tcdrain(file_descriptor);
		if (status < 0)
		{
			return EERRNO;
		}
		status = tcflush(file_descriptor, TCIOFLUSH);
		if (status < 0)
		{
			return EERRNO;
		}
		status = tcsetattr(file_descriptor, TCSANOW, &terminal);
		if (status < 0)
		{
			return EERRNO;
		}
#ifdef DEBUG
		status = tcgetattr(file_descriptor, &terminal);
		if (status < 0)
		{
			return EERRNO;
		}
		ispeed = cfgetispeed(&terminal);
		ospeed = cfgetospeed(&terminal);
		printf("DEBUG: I/O speed: %s/%s (%o/%o) baud, %d ds\n", tospeedstring(ispeed), tospeedstring(ospeed), ispeed, ospeed, terminal.c_cc[VTIME]);
#endif
		return EOK;
	}
	return EAOPENED;
}

vrhi_status_t vrhi_close( void )
{
	if (file_descriptor >= 0)
	{
		ssize_t result;

		result = close(file_descriptor);
		if (result < 0)
		{
			return EERRNO;
		}
		file_descriptor = FDNONE;
		return EOK;
	}
	return ENOPENED;
}

static unsigned char calculate_checksum( unsigned int count )
{
	unsigned short sum;
	unsigned int i;

	sum = 0;
	for (i = 0; i < count; i++)
	{
		sum += (unsigned char)communication_buffer[i];
	}
	return sum + (sum >> 8);
}

static vrhi_status_t send_frame( unsigned char command, unsigned int count )
{
	unsigned char checksum;
	ssize_t result;

	checksum = calculate_checksum(count);

#ifdef DEBUG
	printf("DEBUG: '%c' $%02x\n", command, command);
	if (count > 0)
	{
		printf_buffer("DEBUG: send", count);
		printf("DEBUG: checksum $%02x\n", checksum);
	}
#endif

	result = block_write(&command, 1);
	if (result < 0)
	{
		return EERRNO;
	}
	if (result != 1)
	{
#ifdef DEBUG
		printf("DEBUG: Sent '%c' comand %ld/%d\n", command, result, 1);
#endif
		return ESTOOFEW;
	}

	if (count > 0)
	{
		result = block_write(communication_buffer, count);
		if (result < 0)
		{
			return EERRNO;
		}
		if (result != count)
		{
#ifdef DEBUG
			printf("DEBUG: Sent '%c' body %ld/%u\n", command, result, count);
			printf_buffer("DEBUG: sent", result);
#endif
			return ESTOOFEW;
		}

		result = block_write(&checksum, 1);
		if (result < 0)
		{
			return EERRNO;
		}
		if (result != 1)
		{
#ifdef DEBUG
			printf("DEBUG: Sent '%c' checksum %ld/%d\n", command, result, 1);
#endif
			return ESTOOFEW;
		}
	}

	return EOK;
}

static vrhi_status_t receive_frame(
	unsigned char command, 
	unsigned int count, bool_t check )
{
	unsigned char checksum;
	ssize_t result;
	ssize_t size;

	size = check ? count+1 : count;

	result = block_read(communication_buffer, size);
	if (result < 0)
	{
		return EERRNO;
	}
#ifdef DEBUG
	printf("DEBUG: Received '%c' body %ld/%ld\n", command, result, size);
	printf_buffer("DEBUG: receive", result);
#endif
	if (result != size)
	{
		return ERTOOFEW;
	}

	if (check)
	{
		checksum = calculate_checksum(count);
		if ((unsigned char)communication_buffer[count] != checksum)
		{
#ifdef DEBUG
			printf("DEBUG: Command '%c' checksum $%02x ($%02x)\n", command, (unsigned char)communication_buffer[count], checksum);
#endif
			return EWCSUM;
		}
	}

	return EOK;
}

vrhi_status_t vrhi_reset( void )
{
	if (file_descriptor >= 0)
	{
		vrhi_status_t result;

		result = send_frame('\0', 0);
		if (result != EOK)
		{
			return result;
		}

		result = receive_frame('\0', 1, FALSE);
		if (result != EOK)
		{
			return result;
		}
		if (communication_buffer[0] != 0x00)
		{
			return ENEXISTS;
		}
		return EOK;
	}
	return ENOPENED;
}

vrhi_status_t vrhi_info( vrhi_response_info_t *response )
{
	if (file_descriptor >= 0)
	{
		ssize_t result;

		result = send_frame('?', 0);
		if (result != EOK)
		{
			return result;
		}

		result = receive_frame('?', 4+2+2, TRUE);
		if (result != EOK)
		{
			return result;
		}
		if (strncmp(communication_buffer+0, SIGNATURE, sizeof(SIGNATURE)) != 0)
		{
			return EWSIGNATURE;
		}
		if (strncmp(communication_buffer+4, PLATFORM, sizeof(PLATFORM)) != 0)
		{
			return EWPLATFORM;
		}
		memcpy(response->signature, communication_buffer+0, sizeof(SIGNATURE));
		memcpy(response->platform, communication_buffer+4, sizeof(PLATFORM));
		response->version = communication_buffer[6];
		response->revision = communication_buffer[7];

		return EOK;
	}
	return ENOPENED;
}

vrhi_status_t vrhi_power( vrhi_request_power_t *request, vrhi_response_power_t *response )
{
	if (file_descriptor >= 0)
	{
		ssize_t result;

		communication_buffer[0] = 0
			| ((request->_vpp ? 1 : 0) << 7)
			;
		result = send_frame('P', 1);
		if (result != EOK)
		{
			return result;
		}

		result = receive_frame('P', 1, TRUE);
		if (result != EOK)
		{
			return result;
		}
		response->_vpp = ((communication_buffer[0] >> 7) & 0x01) != 0 ? HIGH : LOW;

		return EOK;
	}
	return ENOPENED;
}

vrhi_status_t vrhi_signals( vrhi_request_signals_t *request, vrhi_response_signals_t *response )
{
	if (file_descriptor >= 0)
	{
		ssize_t result;

		communication_buffer[0] = (request->address >> 0) & 0xff;
		communication_buffer[1] = (request->address >> 8) & 0xff;
		communication_buffer[2] = 0
			| ((request->phi2 ? 1 : 0) << 0)
			| ((request->rw ? 1 : 0) << 1)
			| ((request->rwlate ? 1 : 0) << 2)
			| ((request->_cctl ? 1 : 0) << 3)
			| ((request->_s4 ? 1 : 0) << 4)
			| ((request->_s5 ? 1 : 0) << 5)

			| ((request->_ref ? 1 : 0) << 6)
			| ((request->_d1xx ? 1 : 0) << 7)
			;
		communication_buffer[3] = 0
			/*
			| ((request-> ? 1 : 0) << 0)
			| ((request-> ? 1 : 0) << 1)
			*/
			| ((request->_rdy ? 1 : 0) << 2)
			| ((request->_cas ? 1 : 0) << 3)
			| ((request->_ras ? 1 : 0) << 4)
			| ((request->_latchrw ? 1 : 0) << 5)
			| ((request->_reset ? 1 : 0) << 6)
			| ((request->_halt ? 1 : 0) << 7)
			;
		communication_buffer[4] = request->data;
		
		result = send_frame('S', 5);
		if (result != EOK)
		{
			return result;
		}

		result = receive_frame('S', 4, TRUE);
		if (result != EOK)
		{
			return result;
		}

		response->data = communication_buffer[0];

		response->rd4 = ((communication_buffer[1] >> 0) & 0x01) != 0 ? HIGH : LOW;
		response->rd5 = ((communication_buffer[1] >> 1) & 0x01) != 0 ? HIGH : LOW;

		response->_irq = ((communication_buffer[1] >> 2) & 0x01) != 0 ? HIGH : LOW;
		response->_extsel = ((communication_buffer[1] >> 3) & 0x01) != 0 ? HIGH : LOW;
		response->_mpd = ((communication_buffer[1] >> 4) & 0x01) != 0 ? HIGH : LOW;
		response->_casinh = ((communication_buffer[1] >> 5) & 0x01) != 0 ? HIGH : LOW;
		response->_reset = ((communication_buffer[1] >> 6) & 0x01) != 0 ? HIGH : LOW;
		response->_halt = ((communication_buffer[1] >> 7) & 0x01) != 0 ? HIGH : LOW;

		response->audio = 0
			| ((unsigned short)communication_buffer[2] << 0) 
			| ((unsigned short)communication_buffer[3] << 8)
			;

		return EOK;
	}
	return ENOPENED;
}

vrhi_status_t vrhi_read( vrhi_request_read_t *request, vrhi_response_read_t *response )
{
	if (file_descriptor >= 0)
	{
		ssize_t result;

		communication_buffer[0] = (request->address >> 0) & 0xff;
		communication_buffer[1] = (request->address >> 8) & 0xff;
		communication_buffer[2] = 0
			/*
			| ((request->phi2 ? 1 : 0) << 0)
			| ((request->rw ? 1 : 0) << 1)
			| ((request->rwlate ? 1 : 0) << 2)
			*/
			| ((request->_cctl ? 1 : 0) << 3)
			| ((request->_s4 ? 1 : 0) << 4)
			| ((request->_s5 ? 1 : 0) << 5)

			| ((request->_ref ? 1 : 0) << 6)
			| ((request->_d1xx ? 1 : 0) << 7)
			;
		communication_buffer[3] = 0
			/*
			| ((request-> ? 1 : 0) << 0)
			| ((request-> ? 1 : 0) << 1)
			*/
			| ((request->_rdy ? 1 : 0) << 2)
			| ((request->_cas ? 1 : 0) << 3)
			| ((request->_ras ? 1 : 0) << 4)
			/*
			| ((request->_latchrw ? 1 : 0) << 5)
			*/
			| ((request->_reset ? 1 : 0) << 6)
			| ((request->_halt ? 1 : 0) << 7)
			;
		
		result = send_frame('R', 4);
		if (result != EOK)
		{
			return result;
		}

		result = receive_frame('R', 4, TRUE);
		if (result != EOK)
		{
			return result;
		}

		response->data = communication_buffer[0];

		response->rd4 = ((communication_buffer[1] >> 0) & 0x01) != 0 ? HIGH : LOW;
		response->rd5 = ((communication_buffer[1] >> 1) & 0x01) != 0 ? HIGH : LOW;

		response->_irq = ((communication_buffer[1] >> 2) & 0x01) != 0 ? HIGH : LOW;
		response->_extsel = ((communication_buffer[1] >> 3) & 0x01) != 0 ? HIGH : LOW;
		response->_mpd = ((communication_buffer[1] >> 4) & 0x01) != 0 ? HIGH : LOW;
		response->_casinh = ((communication_buffer[1] >> 5) & 0x01) != 0 ? HIGH : LOW;
		response->_reset = ((communication_buffer[1] >> 6) & 0x01) != 0 ? HIGH : LOW;
		response->_halt = ((communication_buffer[1] >> 7) & 0x01) != 0 ? HIGH : LOW;

		response->audio = 0
			| ((unsigned short)communication_buffer[2] << 0) 
			| ((unsigned short)communication_buffer[3] << 8)
			;

		return EOK;
	}
	return ENOPENED;
}

vrhi_status_t vrhi_write( vrhi_request_write_t *request, vrhi_response_write_t *response )
{
	if (file_descriptor >= 0)
	{
		ssize_t result;

		communication_buffer[0] = (request->address >> 0) & 0xff;
		communication_buffer[1] = (request->address >> 8) & 0xff;
		communication_buffer[2] = 0
			/*
			| ((request->phi2 ? 1 : 0) << 0)
			| ((request->rw ? 1 : 0) << 1)
			| ((request->rwlate ? 1 : 0) << 2)
			*/
			| ((request->_cctl ? 1 : 0) << 3)
			| ((request->_s4 ? 1 : 0) << 4)
			| ((request->_s5 ? 1 : 0) << 5)

			| ((request->_ref ? 1 : 0) << 6)
			| ((request->_d1xx ? 1 : 0) << 7)
			;
		communication_buffer[3] = 0
			/*
			| ((request-> ? 1 : 0) << 0)
			| ((request-> ? 1 : 0) << 1)
			*/
			| ((request->_rdy ? 1 : 0) << 2)
			| ((request->_cas ? 1 : 0) << 3)
			| ((request->_ras ? 1 : 0) << 4)
			/*
			| ((request->_latchrw ? 1 : 0) << 5)
			*/
			| ((request->_reset ? 1 : 0) << 6)
			| ((request->_halt ? 1 : 0) << 7)
			;
		communication_buffer[4] = request->data;
		
		result = send_frame('W', 5);
		if (result != EOK)
		{
			return result;
		}

		result = receive_frame('W', 3, TRUE);
		if (result != EOK)
		{
			return result;
		}

		response->rd4 = ((communication_buffer[0] >> 0) & 0x01) != 0 ? HIGH : LOW;
		response->rd5 = ((communication_buffer[0] >> 1) & 0x01) != 0 ? HIGH : LOW;

		response->_irq = ((communication_buffer[0] >> 2) & 0x01) != 0 ? HIGH : LOW;
		response->_extsel = ((communication_buffer[0] >> 3) & 0x01) != 0 ? HIGH : LOW;
		response->_mpd = ((communication_buffer[0] >> 4) & 0x01) != 0 ? HIGH : LOW;
		response->_casinh = ((communication_buffer[0] >> 5) & 0x01) != 0 ? HIGH : LOW;
		response->_reset = ((communication_buffer[0] >> 6) & 0x01) != 0 ? HIGH : LOW;
		response->_halt = ((communication_buffer[0] >> 7) & 0x01) != 0 ? HIGH : LOW;

		response->audio = 0
			| ((unsigned short)communication_buffer[1] << 0) 
			| ((unsigned short)communication_buffer[2] << 8)
			;

		return EOK;
	}
	return ENOPENED;
}

vrhi_status_t vrhi_chip_info( vrhi_request_chip_info_t *request, vrhi_response_chip_info_t *response )
{
	if (file_descriptor >= 0)
	{
		ssize_t result;

		communication_buffer[0] = (request->address >> 0) & 0xff;
		communication_buffer[1] = (request->address >> 8) & 0xff;
		communication_buffer[2] = (request->address >> 16) & 0xff;

		result = send_frame('i', 3);
		if (result != EOK)
		{
			return result;
		}

		result = receive_frame('i', 2, TRUE);
		if (result != EOK)
		{
			return result;
		}

		response->manufacturer_id = communication_buffer[0];
		response->device_id = communication_buffer[1];

		return EOK;
	}
	return ENOPENED;
}

vrhi_status_t vrhi_chip_erase( vrhi_request_chip_erase_t *request )
{
	if (file_descriptor >= 0)
	{
		ssize_t result;

		communication_buffer[0] = (request->address >> 0) & 0xff;
		communication_buffer[1] = (request->address >> 8) & 0xff;
		communication_buffer[2] = (request->address >> 16) & 0xff;

		result = send_frame('c', 3);
		if (result != EOK)
		{
			return result;
		}

		result = receive_frame('c', 1, FALSE);
		if (result != EOK)
		{
			return result;
		}
		if (communication_buffer[0] != 0x00)
		{
			return ENEXISTS;
		}

		return EOK;
	}
	return ENOPENED;
}

vrhi_status_t vrhi_sector_erase( vrhi_request_sector_erase_t *request )
{
	if (file_descriptor >= 0)
	{
		ssize_t result;

		communication_buffer[0] = (request->address >> 0) & 0xff;
		communication_buffer[1] = (request->address >> 8) & 0xff;
		communication_buffer[2] = (request->address >> 16) & 0xff;

		result = send_frame('e', 3);
		if (result != EOK)
		{
			return result;
		}

		result = receive_frame('e', 1, FALSE);
		if (result != EOK)
		{
			return result;
		}
		if (communication_buffer[0] != 0x00)
		{
			return ENEXISTS;
		}

		return EOK;
	}
	return ENOPENED;
}

vrhi_status_t vrhi_bank_read( vrhi_request_bank_read_t *request, vrhi_response_bank_read_t *response )
{
	if (file_descriptor >= 0)
	{
		ssize_t result;

		communication_buffer[0] = (request->bank >> 0) & 0xff;
		
		result = send_frame('r', 1);
		if (result != EOK)
		{
			return result;
		}

		result = receive_frame('r', BANK_SIZE, TRUE);
		if (result != EOK)
		{
			return result;
		}

		memcpy(response->data, communication_buffer, BANK_SIZE);

		return EOK;
	}
	return ENOPENED;
}

vrhi_status_t vrhi_bank_write( vrhi_request_bank_write_t *request )
{
	if (file_descriptor >= 0)
	{
		ssize_t result;

		communication_buffer[0] = (request->bank >> 0) & 0xff;
		memcpy(communication_buffer+1, request->data, BANK_SIZE);
		
		result = send_frame('w', 1+BANK_SIZE);
		if (result != EOK)
		{
			return result;
		}

		result = receive_frame('w', 1, FALSE);
		if (result != EOK)
		{
			return result;
		}
		if (communication_buffer[0] != 0x00)
		{
			return ENEXISTS;
		}
		return EOK;
	}
	return ENOPENED;
}

