#ifndef _VRHI
#define _VRHI

/*
 * Virtual to Real Hardware Interface.
 * (c) 2024 by Jerzy Kut <mono@protolabo.pl>
 *
 */

#define BANK_SIZE 0x2000

/* Command INFO */

typedef struct
{
	char signature[4];
	char platform[2];
	unsigned char version;
	unsigned char revision;
}
vrhi_response_info_t;

/* Command POWER */

typedef struct
{
	unsigned char _vpp;
}
vrhi_request_power_t;

typedef struct
{
	unsigned char _vpp;
}
vrhi_response_power_t;

/* Command SIGNAL */

typedef struct
{
	unsigned short address;

	unsigned char phi2;
	unsigned char rw;
	unsigned char rwlate;

	unsigned char _cctl;
	unsigned char _s4;
	unsigned char _s5;

	unsigned char _ref;
	unsigned char _d1xx;
	unsigned char _rdy;
	unsigned char _cas;
	unsigned char _ras;

	unsigned char _latchrw;

	unsigned char _reset;
	unsigned char _halt;

	unsigned char data;
}
vrhi_request_signals_t;

typedef struct
{
	unsigned char data;

	unsigned char rd4;
	unsigned char rd5;

	unsigned char _irq;
	unsigned char _extsel;
	unsigned char _mpd;
	unsigned char _casinh;
	unsigned char _reset;
	unsigned char _halt;

	unsigned short audio;
}
vrhi_response_signals_t;

/* Command READ */

typedef struct
{
	unsigned short address;

	/*
	unsigned char phi2;
	unsigned char rw;
	unsigned char rwlate;
	*/
	unsigned char _cctl;
	unsigned char _s4;
	unsigned char _s5;

	unsigned char _ref;
	unsigned char _d1xx;
	unsigned char _rdy;
	unsigned char _cas;
	unsigned char _ras;
	/*
	unsigned char _latchrw;
	*/
	unsigned char _reset;
	unsigned char _halt;
}
vrhi_request_read_t;

typedef struct
{
	unsigned char data;

	unsigned char rd4;
	unsigned char rd5;

	unsigned char _irq;
	unsigned char _extsel;
	unsigned char _mpd;
	unsigned char _casinh;
	unsigned char _reset;
	unsigned char _halt;

	unsigned short audio;
}
vrhi_response_read_t;

/* Command WRITE */

typedef struct
{
	unsigned short address;

	/*
	unsigned char phi2;
	unsigned char rw;
	unsigned char rwlate;
	*/
	unsigned char _cctl;
	unsigned char _s4;
	unsigned char _s5;

	unsigned char _ref;
	unsigned char _d1xx;
	unsigned char _rdy;
	unsigned char _cas;
	unsigned char _ras;
	/*
	unsigned char _latchrw;
	*/
	unsigned char _reset;
	unsigned char _halt;

	unsigned char data;
}
vrhi_request_write_t;

typedef struct
{
	unsigned char rd4;
	unsigned char rd5;

	unsigned char _irq;
	unsigned char _extsel;
	unsigned char _mpd;
	unsigned char _casinh;
	unsigned char _reset;
	unsigned char _halt;

	unsigned short audio;
}
vrhi_response_write_t;

/* Commmand CHIP INFO */

typedef struct
{
	unsigned int address;
}
vrhi_request_chip_info_t;

typedef struct
{
	unsigned char manufacturer_id;
	unsigned char device_id;
}
vrhi_response_chip_info_t;

/* Commmand CHIP ERASE */

typedef struct
{
	unsigned int address;
}
vrhi_request_chip_erase_t;

/* Commmand SECTOR ERASE */

typedef struct
{
	unsigned int address;
}
vrhi_request_sector_erase_t;

/* Commmand BANK READ */

typedef struct
{
	unsigned char bank;
}
vrhi_request_bank_read_t;

typedef struct
{
	char *data;
}
vrhi_response_bank_read_t;

/* Commmand BANK WRITE */

typedef struct
{
	unsigned char bank;
	char *data;
}
vrhi_request_bank_write_t;

/* Status of */
typedef enum
{
	EWPLATFORM = -9,	/* wrong device platform */
	EWSIGNATURE = -8,	/* wrong device signature */
	EWCSUM = -7,		/* wrong checksum */
	ENEXISTS = -6,		/* device not exists */
	ERTOOFEW = -5,		/* too few data received */
	ESTOOFEW = -4,		/* too few data sent */
	ENOPENED = -3,		/* device not opened */
	EAOPENED = -2,		/* device already opened */
	EERRNO = -1,		/* error is in errno */
	EOK = 0			/* all went correctly */
}
vrhi_status_t;

/* Common signal state. */
typedef enum
{
	LOW = 0,	/* "0" */
	HIGH = 1,	/* "1" */
	HI = 1		/* high impedance */
}
vrhi_low_high_t;

/* R/W signal state. */
typedef enum
{
	WRITE = 0,	/* _W */
	READ = 1	/* R */
}
vrhi_read_write_t;

/* Version of the API. */
extern const int vrhi_version;
extern const int vrhi_revision;

/* API */

/**
 * Open the connection.
 * @param serial name of the device file e.g. "/dev/ttyUSB0"
 * @returns status of operation
 * EOK -ok
 * EERRNO -error of operation is in system errno variable
 * EAOPENED -port already opened
 */
vrhi_status_t vrhi_open( char *serial );

/**
 * Close the connection.
 * @returns status of operation
 * EOK -ok
 * EERRNO -error of operation is in system errno variable
 * ENOPENED -port not opened
 */
vrhi_status_t vrhi_close( void );

/**
 * Reset device.
 * This performs initialization of the device
 * in the following sequence of operations:
 * 1. configure data bus to input floating
 * 2. set address bus to all high ($FFFF)
 * 3. set PHI2, R/W, /CCTL, /S4, /S5 and /REF. /D1XX, /RDY, /CAS, /RAS, /RESET, /HALT to high
 * 4. power off the device (/Vpp set to high)
 * @returns status of operation
 * EOK -ok
 * EERRNO -error of operation is in system errno variable
 * ETOOFEW -too few data received
 * ENEXISTS -device not exists
 * ENOPENED -port not opened
 */
vrhi_status_t vrhi_reset( void );

/**
 * Take information about the device.
 * @param response information structure
 * @returns status of operation
 * EOK -ok
 * EERRNO -error of operation is in system errno variable
 * ETOOFEW -too few data received
 * ENOPENED -port not opened
 * EWCSUM -wrong checksum
 * EWSIGNATURE -wrong VRHI signature
 * EWPLATFORM -platform is not A8
 */
vrhi_status_t vrhi_info( vrhi_response_info_t *response );

/**
 * Power configuration on the device.
 * This configure /Vpp power signal on the device.
 * @param request power signals that are set on the operation
 * @param response power signals that are read on the operation
 * @returns status of operation
 * EOK -ok
 * EERRNO -error of operation is in system errno variable
 * ETOOFEW -too few data received
 * ENOPENED -port not opened
 * EWCSUM -wrong checksum
 */
vrhi_status_t vrhi_power( vrhi_request_power_t *request, vrhi_response_power_t *response );

/**
 * Set and read signals operation on the device.
 * This performs set and read read operation on the device
 * in the following sequence of operations:
 * 1. set address bus and data bus
 * 2. configure data bus to input pull-up
 * 3. set PHI2, RW, /CCTL, /S4, /S5 and /REF. /D1XX, /RDY, /CAS, /RAS, /RESET, /HALT
 * 4. read data bus and RD4, RD5 and /IRQ, /EXTSEL, /MPD, /CASINH, /RESET, /HALT
 * 5. configure data bus to input floating
 * @param request signals that are set on the operation
 * @param response signals that are read on the operation
 * @returns status of operation
 * EOK -ok
 * EERRNO -error of operation is in system errno variable
 * ETOOFEW -too few data received
 * ENOPENED -port not opened
 * EWCSUM -wrong checksum
 */
vrhi_status_t vrhi_signals( vrhi_request_signals_t *request, vrhi_response_signals_t *response );

/**
 * Read operation on the device.
 * This performs read operation from the device
 * in the following sequence of operations:
 * 1. push PHI2 to high
 * 2. set R/W to high (READ)
 * 3. set address bus
 * 4. configure data bus to input pull-up
 * 5. set /CCTL, /S4, /S5 and /REF. /D1XX, /RDY, /CAS, /RAS, /RESET, /HALT
 * 6. pull PHI2 to low
 * 7. read data bus and RD4, RD5 and /IRQ, /EXTSEL, /MPD, /CASINH, /RESET, /HALT
 * 8. push PHI2 to high
 * 9. configure data bus to input floating
 * @param request signals that are set on the operation
 * @param response signals that are read on the operation
 * @returns status of operation
 * EOK -ok
 * EERRNO -error of operation is in system errno variable
 * ETOOFEW -too few data received
 * ENOPENED -port not opened
 * EWCSUM -wrong checksum
 */
vrhi_status_t vrhi_read( vrhi_request_read_t *request, vrhi_response_read_t *response );

/**
 * Write operation on the device.
 * This performs write operation on the device
 * in the following sequence of operations:
 * 1. push PHI2 to high
 * 2. set R/W to low (WRITE)
 * 3. set address bus
 * 4. configure data bus to output push-pull
 * 5. set /CCTL, /S4, /S5 and /REF. /D1XX, /RDY, /CAS, /RAS, /RESET, /HALT
 * 6. write value to data bus
 * 7. pull PHI2 to low
 * 8. read RD4, RD5 and /IRQ, /EXTSEL, /MPD, /CASINH, /RESET, /HALT
 * 9. push PHI2 to high
 * 10. set R/W to high (READ)
 * 11. configure data bus to input floating
 * @param request signals that are set on the operation
 * @param response signals that are read on the operation
 * @returns status of operation
 * EOK -ok
 * EERRNO -error of operation is in system errno variable
 * ETOOFEW -too few data received
 * ENOPENED -port not opened
 * EWCSUM -wrong checksum
 */
vrhi_status_t vrhi_write( vrhi_request_write_t *request, vrhi_response_write_t *response );

/**
 * Read memory chip signature.
 * @param request address of chip memory
 * @param response signature details
 * @returns status of operation
 * EOK -ok
 * EERRNO -error of operation is in system errno variable
 * ETOOFEW -too few data received
 * ENEXISTS -device not exists
 * ENOPENED -port not opened
 */
vrhi_status_t vrhi_chip_info( vrhi_request_chip_info_t *request, vrhi_response_chip_info_t *response );

/**
 * Erase chip.
 * @param request address of chip memory
 * @returns status of operation
 * EOK -ok
 * EERRNO -error of operation is in system errno variable
 * ETOOFEW -too few data received
 * ENEXISTS -device not exists
 * ENOPENED -port not opened
 */
vrhi_status_t vrhi_chip_erase( vrhi_request_chip_erase_t *request );

/**
 * Erase chip sector.
 * @param request address of chip sector memory
 * @returns status of operation
 * EOK -ok
 * EERRNO -error of operation is in system errno variable
 * ETOOFEW -too few data received
 * ENEXISTS -device not exists
 * ENOPENED -port not opened
 */
vrhi_status_t vrhi_sector_erase( vrhi_request_sector_erase_t *request );

/**
 * Read memory bank.
 * @param request bank number
 * @param response data
 * @returns status of operation
 * EOK -ok
 * EERRNO -error of operation is in system errno variable
 * ETOOFEW -too few data received
 * ENEXISTS -device not exists
 * ENOPENED -port not opened
 */
vrhi_status_t vrhi_bank_read( vrhi_request_bank_read_t *request, vrhi_response_bank_read_t *response );

/**
 * Write memory bank.
 * @param request bank number and data to write
 * @returns status of operation
 * EOK -ok
 * EERRNO -error of operation is in system errno variable
 * ETOOFEW -too few data received
 * ENEXISTS -device not exists
 * ENOPENED -port not opened
 */
vrhi_status_t vrhi_bank_write( vrhi_request_bank_write_t *request );

#endif /* _VRHI */
