/********************************** (C) COPYRIGHT *******************************
* File Name          : main.c
* Author             : WCH
* Version            : V1.0.0
* Date               : 2021/06/06
* Description        : Main program body.
* Copyright (c) 2021 Nanjing Qinheng Microelectronics Co., Ltd.
* SPDX-License-Identifier: Apache-2.0
*******************************************************************************/

/*
 *@Note
 GPIO big-bang access to Atari's XE CART/ECI bus:

 Changelog:
 2024/04/07: RD4 and RD5 pin assignment exchanged
*/

#define PHI2_DELAY
//#define LONG_SOFTWARE_ID_EXIT

#define SERIAL_TIMEOUT_NS 1000000
#define BANK_SIZE 0x2000

typedef enum
{
	FALSE = 0,
	TRUE = !FALSE
}
bool_t;

/*
 * Pin assignment
 * --------------
 * PA2:		debug Tx
 * PA3:		debug Rx	;reserved for future
 *
 * PA9:		serial Tx
 * PA10:	serial Rx
 * PA11:	serial CTS
 * PA12:	serial RTS
 */
#define DEBUG_PORT GPIOA
#define SERIAL_PORT GPIOA

/*
 * PC0:     AUDIO (IN)
 *
 * PC1:		/Vpp + /5V600XL + /5V1090 (OUT)
 *
 * PC2:		/CAS (OUT)
 * PC3:		/RAS (OUT)
 * PC4:		/CASINH /EXTENB (IN)
 * PC5:
 * PC6:     /HALT (IN/OUT)
 * PC7:     /RESET (OUT) IN/OUT
 * PC8:     /IRQ (IN)
 * PC9:		/EXTSEL (IN)
 * PC10:	/MPD (IN)
 * PC11:	/REF (OUT)
 * PC12:    /D1XX (OUT)
 * PC13:
 * PC14:
 * PC15:
 *
 * PD0-PD7: DataBus (IN/OUT)
 *
 * PD8:     PHI2 (OUT)
 * PD9:	    R/W (OUT)
 * PD10:    /CCTL (OUT)
 * PD11:    /RDY (OUT)
 * PD12:    /S5 (OUT)
 * PD13:    RD5 (IN)	;później RD4
 * PD14:    /S4 (OUT)
 * PD15:    RD4 (IN)	;później RD5
 *
 * PE0-PE15 Address Bus (OUT)
 *
 * not mapped:
 * RWLATE (OUT) on PBI
 * LATCHRW (OUT) on right cartridge
 *
 */
typedef enum
{
	/*PD*/
	_RDY = 11,

	/*PC*/
	AUDIO = 0,
	_CAS = 2,
	_RAS = 3,
	_CASINH = 4,
	_HALT = 6,
	_RESET = 7,
	_IRQ = 8,
	_EXTSEL = 9,
	_MPD = 10,
	_REF = 11,
	_D1XX= 12,
}
parallel_bus_control_t;

typedef enum
{
	/*PC*/
	_VPP = 1
}
voltage_control_t;

typedef enum
{
	/*PD*/
	PHI2 = 8,
	RW = 9,
	_CCTL = 10,
	_S5 = 12,
	RD5 = 13,	/* w najświeższej wersji RD4 */
	_S4 = 14,
	RD4 = 15	/* w najświeższej wersji RD5 */
}
cartridge_control_t;

#define VOLTAGE_PORT GPIOC
#define PARALLEL_PORT GPIOC
#define DATA_PORT GPIOD
#define CARTRIDGE_PORT GPIOD
#define ADDRESS_PORT GPIOE

/*
 * Protocol
 * --------
 *
 * INFO
 *
 * request:
 * $00: $3F - '?'
 *
 * response:
 * $00: $56 - 'V'
 * $01: $52 - 'R'
 * $02: $48 - 'H'
 * $03: $49 - 'I'
 * $04: $41 - 'A' atari 800
 * $05: $38 - '8'
 * $06: $01 - hardware version
 * $07: $00 - hardware revision
 * $08: $xx - checksum of bytes $00..$07
 *
 *
 * POWER
 *
 * request:
 * $00: $50       - 'P'
 * $01: %V....... - 7 - /Vpp
 * $02: $xx       - checksum of bytes $01..$01
 *
 * response:
 * $00: %V....... - 7 - /Vpp
 * $01: $xx       - checksum of bytes $00..$00
 *
 *
 * SIGNALS
 *
 * request:
 * $00: $53       - 'S'
 * $01: %aaaaaaaa - A0..A7 - address LSB
 * $02: %aaaaaaaa - A8..A15 - address MSB
 * $03: %1RSsCLW2 - 2 - PHI2
 *                  W - R/W
 *                  L - R/WLATE (?)
 *                  C - /CCTL
 *                  s - /S4
 *                  S - /S5
 *                  R - /REF
 *                  1 - /D1XX
 * $04: %H0LRCY.. - Y - /RDY
 *                  C - /CAS
 *                  R - /RAS
 *                  L - LATCHR/W (?)
 *                  0 - /RESET
 *                  H - /HALT
 * $05: %dddddddd - data
 * $06: $xx       - checksum of bytes $01..$05
 *
 * response:
 * $00: %dddddddd - D0-D7 - data
 * $01: %H0IMEiRr - r - RD4
 *                  R - RD5
 *                  I - /IRQ
 *                  E - /EXTSEL
 *                  M - /MPD
 *                  i - /CASINH
 *                  0 - /RESET
 *                  H - /HALT
 * $02: %aaaa.... - audio lsb
 * $03: %aaaaaaaa - audio msb
 * $04: $xx       - checksum of bytes $00..$03
 *
 *
 * READ
 *
 * request:
 * $00: $52       - 'R'
 * $01: %aaaaaaaa - A0..A7 - address LSB
 * $02: %aaaaaaaa - A8..A15 - address MSB
 * $03: %1RSsC... - C - /CCTL
 *                  s - /S4
 *                  S - /S5
 *                  R - /REF
 *                  1 - /D1XX
 * $04: %H0LRCY.. - Y - /RDY
 *                  C - /CAS
 *                  R - /RAS
 *                  L - LATCHR/W (?)
 *                  0 - /RESET
 *                  H - /HALT
 * $05: $xx       - checksum of bytes $01..$04
 *
 * response:
 * $00: %dddddddd - D0-D7 - data
 * $01: %H0IMEiRr - r - RD4
 *                  R - RD5
 *                  i - /IRQ
 *                  E - /EXTSEL
 *                  M - /MPD
 *                  I - /CASINH
 *                  0 - /RESET
 *                  H - /HALT
 * $02: %aaaa.... - audio lsb
 * $03: %aaaaaaaa - audio msb
 * $04: $xx       - checksum of bytes $00..$03
 *
 *
 * WRITE
 *
 * request:
 * $00: $57       - 'W'
 * $01: %aaaaaaaa - A0..A7 - address LSB
 * $02: %aaaaaaaa - A8..A12 - address MSB
 * $03: %1RSsC... - C - /CCTL
 *                  s - /S4
 *                  S - /S5
 *                  R - /REF
 *                  1 - /D1XX
 * $04: %H0LRCY.. - Y - /RDY
 *                  C - /CAS
 *                  R - /RAS
 *                  L - LATCHR/W (?)
 *                  0 - /RESET
 *                  H - /HALT
 * $05: %dddddddd - data
 * $06: $xx       - checksum of bytes $01..$05
 *
 * response:
 * $00: %H0IMEiRr - r - RD4
 *                  R - RD5
 *                  I - /IRQ
 *                  E - /EXTSEL
 *                  M - /MPD
 *                  i - /CASINH
 *                  0 - /RESET
 *                  H - /HALT
 * $01: %aaaa.... - audio lsb
 * $02: %aaaaaaaa - audio msb
 * $03: $xx       - checksum of bytes $00..$02
 *
 * -------------------------------------------------
 *
 * CHIP INFO
 *
 * request:
 * $00: $69       - 'i'
 * $01: %bbbbbbbb - A0..A7 - address
 * $02: %bbbbbbbb - A8..A15 - address
 * $03: %bbbbbbbb - A16..A23 - address
 * $04: $xx       - checksum of bytes $01..$03
 *
 * response:
 * $00: %dddddddd - D0-D7 - manufacturer id
 * $01: %dddddddd - D0-D7 - device id
 * $02: $xx       - checksum of bytes $00..$01
 *
 *
 * CHIP ERASE
 *
 * request:
 * $00: $63       - 'c'
 * $01: %bbbbbbbb - A0..A7 - address
 * $02: %bbbbbbbb - A8..A15 - address
 * $03: %bbbbbbbb - A16..A23 - address
 * $04: $xx       - checksum of bytes $01..$03
 *
 * response:
 * $00: $00       - checksum of none
 *
 *
 * SECTOR ERASE
 *
 * request:
 * $00: $65       - 'e'
 * $01: %bbbbbbbb - A0..A7 - address
 * $02: %bbbbbbbb - A8..A15 - address
 * $03: %bbbbbbbb - A16..A23 - address
 * $04: $xx       - checksum of bytes $01..$03
 *
 * response:
 * $00: $00       - checksum of none
 *
 *
 * BANK READ
 *
 * request:
 * $00: $72       - 'r'
 * $01: %bbbbbbbb - B0..B7 - bank
 * $02: $xx       - checksum of bytes $01..$01
 *
 * response:
 * $00: %dddddddd - D0-D7 - data
 * ...
 *  nn: $xx       - checksum of bytes $00..nn-1
 *
 *
 * BANK WRITE
 *
 * request:
 * $00: $77       - 'w'
 * $01: %bbbbbbbb - B0..B7 - bank
 * $02: %dddddddd - data
 * ...
 *  nn: $xx       - checksum of bytes $01..nn-1
 *
 * response:
 * $00: $00       - checksum of none
 */

#include "debug.h"


/* Global typedef */


/* Global define */

#define FW_VERSION 1
#define FW_REVISION 5

/* Global constants */

/* Device information block */
static const u8 INFO[] =
{
		'V', 'R', 'H', 'I',
		'A', '8',
		FW_VERSION, FW_REVISION,
};

/* Global Variable */

/* Data bus (GPIOD) configuration input floating / input pullup / output push */
/* this structure after execution of GPIO_Prepare is configured as following:
 * GPIO_InitStructure.GPIO_Pin = 0
 * 		| GPIO_Pin_0			//D0
 * 		| GPIO_Pin_1			//D1
 * 		| GPIO_Pin_2			//D2
 * 		| GPIO_Pin_3			//D3
 * 		| GPIO_Pin_4			//D4
 * 		| GPIO_Pin_5			//D5
 * 		| GPIO_Pin_6			//D6
 * 		| GPIO_Pin_7			//D7
 * 		;
 * GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN_FLOATING;
 * GPIO_InitStructure.GPIO_Speed = GPIO_Speed_2MHz;
 */
static GPIO_InitTypeDef GPIOD_InitStructure;

/* Voltage control bits */
static u16 voltage;

/* Communication buffer */
static u8 buffer[1+BANK_SIZE+1];

/* ---------------------------------------------- */

#ifdef PHI2_DELAY
/* approximate delay in ns - based on Delay_Us taken from debug.c */
void Delay_Ns(uint32_t n)
{
	uint32_t i;

	SysTick->SR &= ~(1 << 0);

	i = SystemCoreClock / 8000000;
	i = ((i * n) + 999) / 1000;	/* ceil */

	SysTick->CMP = i;
	SysTick->CTLR |= (1 << 4);
	SysTick->CTLR |= (1 << 5) | (1 << 0);

	while((SysTick->SR & (1 << 0)) != (1 << 0))
	{
	}
	SysTick->CTLR &= ~(1 << 0);
}
#endif

void UART_Prepare(void)
{
	GPIO_InitTypeDef GPIO_InitStructure;
	USART_InitTypeDef USART_InitStructure;

	RCC_APB2PeriphClockCmd(RCC_APB2Periph_GPIOA | RCC_APB2Periph_USART1, ENABLE);

	/* USART1 */
	GPIO_InitStructure.GPIO_Pin = 0
			| GPIO_Pin_9	/* TX-->A.9 */
		    | GPIO_Pin_12	/* RTS-->A.12 */
			;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF_PP;
	GPIO_Init(GPIOA, &GPIO_InitStructure);

	GPIO_InitStructure.GPIO_Pin = 0
			| GPIO_Pin_10	/* RX-->A.10 */
		    | GPIO_Pin_11	/* CTS-->A.11 */
			;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN_FLOATING;
	GPIO_Init(GPIOA, &GPIO_InitStructure);

	GPIO_PinLockConfig(GPIOA, 0
			| GPIO_Pin_9	/* TX-->A.9 */
			| GPIO_Pin_10	/* RX-->A.10 */
		    | GPIO_Pin_11	/* CTS-->A.11 */
		    | GPIO_Pin_12	/* RTS-->A.12 */
	);

	/*USART_InitStruct->USART_BaudRate = 9600;
	USART_InitStruct->USART_WordLength = USART_WordLength_8b;
	USART_InitStruct->USART_StopBits = USART_StopBits_1;
	USART_InitStruct->USART_Parity = USART_Parity_No;
	USART_InitStruct->USART_Mode = USART_Mode_Rx | USART_Mode_Tx;
	USART_InitStruct->USART_HardwareFlowControl = USART_HardwareFlowControl_None;*/
	USART_StructInit(&USART_InitStructure);

	USART_InitStructure.USART_BaudRate = 921600;
	USART_Init(USART1, &USART_InitStructure);

	USART_Cmd(USART1, ENABLE);
}

void GPIO_Prepare(void)
{
	ADC_InitTypeDef ADC_InitStructure;
	GPIO_InitTypeDef GPIOC_InitStructure;
	GPIO_InitTypeDef GPIOE_InitStructure;

	RCC_APB2PeriphClockCmd(RCC_APB2Periph_GPIOC, ENABLE);	//PARALLELPORT, VOLTAGEPORT
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_GPIOD, ENABLE);	//DATAPORT, CARTRIDGEPORT
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_GPIOE, ENABLE);	//ADDRPORT

	/*
	 * PE0-PE15 Address Bus
	 * address bus is configured as output push/pull
	 */
	/*GPIOE_InitStructure.GPIO_Pin = GPIO_Pin_All;
	GPIOE_InitStructure.GPIO_Mode = GPIO_Mode_IN_FLOATING;
	GPIOE_InitStructure.GPIO_Speed = GPIO_Speed_2MHz;*/
	GPIO_StructInit(&GPIOE_InitStructure);

	/*GPIOE_InitStructure.GPIO_Pin = GPIO_Pin_All;*/			// A0-A15 (OUT)
	GPIOE_InitStructure.GPIO_Mode = GPIO_Mode_Out_PP;
	/*GPIOE_InitStructure.GPIO_Speed = GPIO_Speed_2MHz;*/
	GPIO_Init(GPIOE, &GPIOE_InitStructure);

	/* lock configuration of address bus */
	GPIO_PinLockConfig(GPIOE, GPIO_Pin_All);

	/*
	 * PC0:     AUDIO (IN)
	 * PC1:		/Vpp + /5V600XL + /5V1090 (OUT)
	 * PC2:		/CAS (OUT)
	 * PC3:		/RAS (OUT)
	 * PC4:		/CASINH /EXTENB (IN)
	 * PC5:
	 * PC6:     /HALT (IN/OUT)
	 * PC7:     /RESET (IN/OUT)
	 * PC8:     /IRQ (IN)
	 * PC9:		/EXTSEL (IN)
	 * PC10:	/MPD (IN)
	 * PC11:	/REF (OUT)
	 * PC12:    /D1XX (OUT)
	 * PC13:
	 * PC14:
	 * PC15:
	 * voltage control line is configured as push/pull
	 * audio line is configured as analog input
	 * parallel bus lines are configured as input pullup or output push/pull
	 * unused lines are input floating
	 */
	/*GPIOC_InitStructure.GPIO_Pin = GPIO_Pin_All;
	GPIOC_InitStructure.GPIO_Mode = GPIO_Mode_IN_FLOATING;
	GPIOC_InitStructure.GPIO_Speed = GPIO_Speed_2MHz;*/
	GPIO_StructInit(&GPIOC_InitStructure);

	GPIOC_InitStructure.GPIO_Pin = 0
			| GPIO_Pin_5			//
			| GPIO_Pin_13			//
			| GPIO_Pin_14			//
			| GPIO_Pin_15			//
			;
	/*GPIOC_InitStructure.GPIO_Mode = GPIO_Mode_IN_FLOATING;*/
	GPIO_Init(GPIOC, &GPIOC_InitStructure);

	GPIOC_InitStructure.GPIO_Pin = 0
			| (1 << _VPP)			// /Vpp + /5V600XL + /5V1090 (OUT)
			| (1 << _CAS)			// /CAS (OUT)
			| (1 << _RAS)			// /RAS (OUT)
			| (1 << _REF)			// /REF (OUT)
			| (1 << _D1XX)			// /D1XX (OUT)
			;
	GPIOC_InitStructure.GPIO_Mode = GPIO_Mode_Out_PP;
	/*GPIOC_InitStructure.GPIO_Speed = GPIO_Speed_2MHz;*/
	GPIO_Init(GPIOC, &GPIOC_InitStructure);

	GPIOC_InitStructure.GPIO_Pin = 0
			| (1 << _HALT)			// /HALT (IN/OUT)
			| (1 << _RESET)			// /RESET (IN/OUT)
			;
	GPIOC_InitStructure.GPIO_Mode = GPIO_Mode_Out_OD;
	/*GPIOC_InitStructure.GPIO_Speed = GPIO_Speed_2MHz;*/
	GPIO_Init(GPIOC, &GPIOC_InitStructure);

	GPIOC_InitStructure.GPIO_Pin = 0
			| (1 << AUDIO)			// AUDIO (IN)
			;
	GPIOC_InitStructure.GPIO_Mode = GPIO_Mode_AIN;
	GPIO_Init(GPIOC, &GPIOC_InitStructure);

	GPIOC_InitStructure.GPIO_Pin = 0
			| (1 << _CASINH)		// /CASINH /EXTENB (IN)
			| (1 << _IRQ)			// /IRQ (IN)
			| (1 << _EXTSEL)		// /EXTSEL (IN)
			| (1 << _MPD)			// /MPD (IN)
			;
	GPIOC_InitStructure.GPIO_Mode = GPIO_Mode_IPU;
	GPIO_Init(GPIOC, &GPIOC_InitStructure);

	/* lock configuration of parallel bus, voltage control and all unused signals */
	GPIO_PinLockConfig(GPIOC, 0
			| (1 << AUDIO)			// AUDIO (IN)
			| (1 << _VPP)			// /Vpp + /5V600XL + /5V1090 (OUT)
			| (1 << _CAS)			// /CAS (OUT)
			| (1 << _RAS)			// /RAS (OUT)
			| (1 << _CASINH)		// /CASINH /EXTENB (IN)
			| GPIO_Pin_5			//
			| (1 << _HALT)			// /HALT (IN/OUT)
			| (1 << _RESET)			// /RESET (IN/OUT)
			| (1 << _IRQ)			// /IRQ (IN)
			| (1 << _EXTSEL)		// /EXTSEL (IN)
			| (1 << _MPD)			// /MPD (IN)
			| (1 << _REF)			// /REF (OUT)
			| (1 << _D1XX)			// /D1XX (OUT)
			| GPIO_Pin_13			//
			| GPIO_Pin_14			//
			| GPIO_Pin_15			//
	);

	/*
	 * PD0-PD7: data bus
	 * PD8:     PHI2
	 * PD9:	    R/W
	 * PD10:    /CCTL
	 * PD11:    /RDY (parallel bus signal)
	 * PD12:    /S5
	 * PD13:    RD5
	 * PD14:    /S4
	 * PD15:    RD4
	 * control bus bits are fixed here as input pull-up or output push/pull
	 */
	/*GPIOD_InitStructure.GPIO_Pin = GPIO_Pin_All;
	GPIOD_InitStructure.GPIO_Mode = GPIO_Mode_IN_FLOATING;
	GPIOD_InitStructure.GPIO_Speed = GPIO_Speed_2MHz;*/
	GPIO_StructInit(&GPIOD_InitStructure);

	GPIOD_InitStructure.GPIO_Pin = 0
			| (1 << PHI2)			// PHI2 (OUT)
			| (1 << RW)				// R/W (OUT)
			| (1 << _CCTL)			// /CCTL (OUT)
			| (1 << _RDY)			// /RDY (OUT)
			| (1 << _S5)			// /S5 (OUT)
			| (1 << _S4)			// /S4 (OUT)
			;
	GPIOD_InitStructure.GPIO_Mode = GPIO_Mode_Out_PP;
	/*GPIOD_InitStructure.GPIO_Speed = GPIO_Speed_2MHz;*/
	GPIO_Init(GPIOD, &GPIOD_InitStructure);

	GPIOD_InitStructure.GPIO_Pin = 0
			| (1 << RD5)			// RD5 (IN)
			| (1 << RD4)			// RD4 (IN)
			;
	GPIOD_InitStructure.GPIO_Mode = GPIO_Mode_IPU;
	GPIO_Init(GPIOD, &GPIOD_InitStructure);

	/* lock configuration of control bus (data bus can be reconfigured freely!) */
	GPIO_PinLockConfig(GPIOD, 0
			| (1 << PHI2)			// PHI2 (OUT)
			| (1 << RW)				// R/W (OUT)
			| (1 << _CCTL)			// /CCTL (OUT)
			| (1 << _RDY)			// /RDY (OUT)
			| (1 << _S5)			// /S5 (OUT)
			| (1 << RD5)			// RD5 (IN)
			| (1 << _S4)			// /S4 (OUT)
			| (1 << RD4)			// RD4 (IN)
	);

	/* ADC on AUDIO line configuration */
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_ADC1, ENABLE );
	RCC_ADCCLKConfig(RCC_PCLK2_Div8);	//9MHz

	/*ADC_InitStruct->ADC_Mode = ADC_Mode_Independent;
	ADC_InitStruct->ADC_ScanConvMode = DISABLE;
	ADC_InitStruct->ADC_ContinuousConvMode = DISABLE;
	ADC_InitStruct->ADC_ExternalTrigConv = ADC_ExternalTrigConv_T1_CC1;
	ADC_InitStruct->ADC_DataAlign = ADC_DataAlign_Right;
	ADC_InitStruct->ADC_NbrOfChannel = 1;*/
	ADC_StructInit(&ADC_InitStructure);

	ADC_DeInit(ADC1);
	ADC_InitStructure.ADC_ContinuousConvMode = ENABLE;
	ADC_InitStructure.ADC_ExternalTrigConv = ADC_ExternalTrigConv_None;
	ADC_InitStructure.ADC_DataAlign = ADC_DataAlign_Left;
	ADC_InitStructure.ADC_OutputBuffer = ADC_OutputBuffer_Disable;
	ADC_InitStructure.ADC_Pga = ADC_Pga_1;
	ADC_Init(ADC1, &ADC_InitStructure);

	ADC_RegularChannelConfig(ADC1, ADC_Channel_1, 1, ADC_SampleTime_71Cycles5);	//125kHz

	ADC_Cmd(ADC1, ENABLE);

	ADC_ResetCalibration(ADC1);
	while(ADC_GetResetCalibrationStatus(ADC1))
	{
	}
	ADC_StartCalibration(ADC1);
	while(ADC_GetCalibrationStatus(ADC1))
	{
	}
}

void PORT_Reset(void)
{
	/*
	 * PD0-PD7: DataBus
	 * data bus configured here is input floating
	 * because it is reconfigured due to the read or write access
	 */
	GPIOD_InitStructure.GPIO_Pin = 0
			| GPIO_Pin_0			//D0 (floating)
			| GPIO_Pin_1			//D1 (floating)
			| GPIO_Pin_2			//D2 (floating)
			| GPIO_Pin_3			//D3 (floating)
			| GPIO_Pin_4			//D4 (floating)
			| GPIO_Pin_5			//D5 (floating)
			| GPIO_Pin_6			//D6 (floating)
			| GPIO_Pin_7			//D7 (floating)
			;
	GPIOD_InitStructure.GPIO_Mode = GPIO_Mode_IN_FLOATING;
	GPIO_Init(GPIOD, &GPIOD_InitStructure);

	/* initialize address bus to all ones */
	GPIO_Write(GPIOE, 0xffff);

	/* initialize data and control bus to all ones except of pfi2 */
	GPIO_Write(GPIOD, 0
			| (0x00ff << 0)			/* D0-D7 (floating) */
			| (0 << PHI2)			/* PHI2 (OUT) */
			| (1 << RW)				/* R/W (OUT) */
			| (1 << _CCTL)			/* /CCTL (OUT) */
			| (1 << _RDY)			/* /RDY (OUT) */
			| (1 << _S5)			/* /S5 (OUT) */
			| (1 << RD5)			/* RD5 (IN) */
			| (1 << _S4)			/* /S4 (OUT) */
			| (1 << RD4)			/* RD4 (IN) */
	);

	/* initialize parallel bus and voltage control to all ones */
	GPIO_Write(GPIOC, 0
			| (1 << AUDIO)			/* PC0: AUDIO */
			| (1 << _VPP)			/* PC1: /Vpp */
			| (1 << _CAS)			/* PC2: /CAS */
			| (1 << _RAS)			/* PC3: /RAS */
			| (1 << _CASINH)		/* PC4: _CASINH */
			| GPIO_Pin_5			/* PC5: */
			| (1 << _HALT)			/* PC6: /HALT */
			| (1 << _RESET)			/* PC7: /RESET */
			| (1 << _IRQ)			/* PC8: _IRQ */
			| (1 << _EXTSEL)		/* PC9: _EXTSEL */
			| (1 << _MPD)			/* PC10: _MPD */
			| (1 << _REF)			/* PC11: /REF */
			| (1 << _D1XX)			/* PC12: /D1XX */
			| GPIO_Pin_13			/* PC13: */
			| GPIO_Pin_14			/* PC14: */
			| GPIO_Pin_15			/* PC15: */
	);
    voltage = (1 << _VPP);	/* power off */
}

u8 PORT_Power(u8 power)
{
	u8 output_vpp, input_voltage_control;

	output_vpp = (power >> 7) & 0x01;

	voltage = output_vpp << _VPP;

	GPIO_WriteBit(GPIOC, 1 << _VPP, output_vpp != 0 ? Bit_SET : Bit_RESET);

#ifdef PHI2_DELAY
	Delay_Ns(200);
#endif

	input_voltage_control = GPIO_ReadInputData(GPIOC);

	power = 0
			| ((input_voltage_control >> _VPP) << 7)	/* PC1: _VPP */
			;
	return power;
}

u32 PORT_Signals(u32 signals, u8 data)
{
	u16 address_bus;
	u16 output_data_and_cartridge_bus;
	u16 output_parallel_and_voltage_bus;
	u32 input_data_and_cartridge_bus;
	u32 input_parallel_and_voltage_bus;
	u32 input_audio;

	address_bus = (signals >> 0) & 0xffff;						/* PE0-PE15: A0-A15 */
	output_data_and_cartridge_bus = 0
			| ((u32)data << 0)									/* PD0-PD7: D0-D7 */
			| (((signals >> 16) & 0x0001) << PHI2)				/* PD8: phi2 */
			| (((signals >> 17) & 0x0001) << RW)				/* PD9: R/W */
			| (((signals >> 19) & 0x0001) << _CCTL)				/* PD10: /CCTL */
			| (((signals >> 26) & 0x0001) << _RDY)				/* PD11: /RDY */
			| (((signals >> 21) & 0x0001) << _S5)				/* PD12: /S5 */
			//| (1 << RD5)										/* PD13: RD5 */
			| (((signals >> 20) & 0x0001) << _S4)				/* PD14: /S4 */
			//| (1 << RD4)										/* PD15: RD4 */
			;
	output_parallel_and_voltage_bus = 0
			//| (1 << AUDIO)									/* PC0: AUDIO */
			| voltage											/* PC1: /Vpp */
			| (((signals >> 27) & 0x0001) << _CAS)				/* PC2: /CAS */
			| (((signals >> 28) & 0x0001) << _RAS)				/* PC3: /RAS */
			//| (1 << _CASINH)									/* PC4: _CASINH */
			//| GPIO_Pin_5										/* PC5: */
			| (((signals >> 31) & 0x0001) << _HALT)				/* PC6: /HALT */
			| (((signals >> 30) & 0x0001) << _RESET)			/* PC7: /RESET */
			//| (1 << _IRQ)										/* PC8: _IRQ */
			//| (1 << _EXTSEL)									/* PC9: _EXTSEL */
			//| (1 << _MPD)										/* PC10: _MPD */
			| (((signals >> 22) & 0x0001) << _REF)				/* PC11: /REF */
			| (((signals >> 23) & 0x0001) << _D1XX)				/* PC12: /D1XX */
			//| GPIO_Pin_13										/* PC13: */
			//| GPIO_Pin_14										/* PC14: */
			//| GPIO_Pin_15										/* PC15: */
			;

	GPIO_Write(GPIOE, address_bus);
	GPIO_Write(GPIOD, output_data_and_cartridge_bus);
	GPIO_Write(GPIOC, output_parallel_and_voltage_bus);
	//printf("PORT_Put %04x:%04x:%04x\r\n", address_bus, output_data_and_cartridge_bus, output_parallel_and_voltage_bus);

	/*GPIOD_InitStructure.GPIO_Pin = 0
			| GPIO_Pin_0			//D0 (IN pullup)
			| GPIO_Pin_1			//D1 (IN pullup)
			| GPIO_Pin_2			//D2 (IN pullup)
			| GPIO_Pin_3			//D3 (IN pullup)
			| GPIO_Pin_4			//D4 (IN pullup)
			| GPIO_Pin_5			//D5 (IN pullup)
			| GPIO_Pin_6			//D6 (IN pullup)
			| GPIO_Pin_7			//D7 (IN pullup)
			;*/
	GPIOD_InitStructure.GPIO_Mode = GPIO_Mode_IPU;
	GPIO_Init(GPIOD, &GPIOD_InitStructure);

#ifdef PHI2_DELAY
	Delay_Ns(200);
#endif

	input_data_and_cartridge_bus = (u32)GPIO_ReadInputData(GPIOD);
	input_parallel_and_voltage_bus = (u32)GPIO_ReadInputData(GPIOC);
	input_audio = (u32)ADC_GetConversionValue(ADC1);
	//printf("PORT_Get %04x:%04x:%04x\r\n", input_data_and_cartridge_bus, input_parallel_and_voltage_bus, input_audio);

	/*GPIOD_InitStructure.GPIO_Pin = 0
			| GPIO_Pin_0			//D0 (IN floating)
			| GPIO_Pin_1			//D1 (IN floating)
			| GPIO_Pin_2			//D2 (IN floating)
			| GPIO_Pin_3			//D3 (IN floating)
			| GPIO_Pin_4			//D4 (IN floating)
			| GPIO_Pin_5			//D5 (IN floating)
			| GPIO_Pin_6			//D6 (IN floating)
			| GPIO_Pin_7			//D7 (IN floating)
			;*/
	GPIOD_InitStructure.GPIO_Mode = GPIO_Mode_IN_FLOATING;
	GPIO_Init(GPIOD, &GPIOD_InitStructure);

	signals = (((input_data_and_cartridge_bus >> 0) & 0x00ff) << 0)				/* PD0-PD7: D0-D7*/
			| (((input_data_and_cartridge_bus >> RD4) & 0x0001) << 8)			/* PD15: RD4 */
			| (((input_data_and_cartridge_bus >> RD5) & 0x0001) << 9)			/* PD13: RD5 */
			| (((input_parallel_and_voltage_bus >> _IRQ) & 0x0001) << 10)		/* PC8: _IRQ */
			| (((input_parallel_and_voltage_bus >> _EXTSEL) & 0x0001) << 11)	/* PC9: _EXTSEL */
			| (((input_parallel_and_voltage_bus >> _MPD) & 0x0001) << 12)		/* PC10: _MPD */
			| (((input_parallel_and_voltage_bus >> _CASINH) & 0x0001) << 13)	/* PC4: _CASINH */
			| (((input_parallel_and_voltage_bus >> _RESET) & 0x0001) << 14)		/* PC7: _RESET */
			| (((input_parallel_and_voltage_bus >> _HALT) & 0x0001) << 15)		/* PC6: _HALT */
			| (((input_audio >> 0) & 0xffff) << 16)								/* ADC4-ADC15: AUDIO */
			;
	return signals;
}

u32 PORT_Read(u32 signals)
{
	u16 address_bus;
	u16 output_data_and_cartridge_bus;
	u16 output_parallel_and_voltage_bus;
	u32 input_data_and_cartridge_bus;
	u32 input_parallel_and_voltage_bus;
	u32 input_audio;

	address_bus = (signals >> 0) & 0xffff;						/* PE0-PE15: A0-A15 */
	output_data_and_cartridge_bus = 0
			//| (0x00ff << 0)									/* PD0-PD7: D0-D7 */
			| (/*((signals >> 16) & 0x0001)*/0 << PHI2)			/* PD8: phi2 */
			| (/*((signals >> 17) & 0x0001)*/1 << RW)			/* PD9: R/W */
			| (((signals >> 19) & 0x0001) << _CCTL)				/* PD10: /CCTL */
			| (((signals >> 26) & 0x0001) << _RDY)				/* PD11: /RDY */
			| (((signals >> 21) & 0x0001) << _S5)				/* PD12: /S5 */
			//| (1 << RD5)										/* PD13: RD5 */
			| (((signals >> 20) & 0x0001) << _S4)				/* PD14: /S4 */
			//| (1 << RD4)										/* PD15: RD4 */
			;
	output_parallel_and_voltage_bus = 0
			//| (1 << AUDIO)									/* PC0: AUDIO */
			| voltage											/* PC1: /Vpp */
			| (((signals >> 27) & 0x0001) << _CAS)				/* PC2: /CAS */
			| (((signals >> 28) & 0x0001) << _RAS)				/* PC3: /RAS */
			//| (1 << _CASINH)									/* PC4: _CASINH */
			//| GPIO_Pin_5										/* PC5: */
			| (((signals >> 31) & 0x0001) << _HALT)				/* PC6: /HALT */
			| (((signals >> 30) & 0x0001) << _RESET)			/* PC7: /RESET */
			//| (1 << _IRQ)										/* PC8: _IRQ */
			//| (1 << _EXTSEL)									/* PC9: _EXTSEL */
			//| (1 << _MPD)										/* PC10: _MPD */
			| (((signals >> 22) & 0x0001) << _REF)				/* PC11: /REF */
			| (((signals >> 23) & 0x0001) << _D1XX)				/* PC12: /D1XX */
			//| GPIO_Pin_13										/* PC13: */
			//| GPIO_Pin_14										/* PC14: */
			//| GPIO_Pin_15										/* PC15: */
			;

	GPIO_Write(GPIOE, address_bus);
	GPIO_Write(GPIOD, output_data_and_cartridge_bus);
	GPIO_Write(GPIOC, output_parallel_and_voltage_bus);
	//printf("PORT_Put %04x:%04x:%04x\r\n", address_bus, output_data_and_cartridge_bus, output_parallel_and_voltage_bus);

	/*GPIOD_InitStructure.GPIO_Pin = 0
			| GPIO_Pin_0			//D0 (IN pullup)
			| GPIO_Pin_1			//D1 (IN pullup)
			| GPIO_Pin_2			//D2 (IN pullup)
			| GPIO_Pin_3			//D3 (IN pullup)
			| GPIO_Pin_4			//D4 (IN pullup)
			| GPIO_Pin_5			//D5 (IN pullup)
			| GPIO_Pin_6			//D6 (IN pullup)
			| GPIO_Pin_7			//D7 (IN pullup)
			;*/
	GPIOD_InitStructure.GPIO_Mode = GPIO_Mode_IPU;
	GPIO_Init(GPIOD, &GPIOD_InitStructure);

#ifdef PHI2_DELAY
	/*Delay_Ns(225);*/
#endif

	GPIO_WriteBit(GPIOD, 1 << PHI2, Bit_SET);

#ifdef PHI2_DELAY
	Delay_Ns(200);
#endif

	GPIO_WriteBit(GPIOD, 1 << PHI2, Bit_RESET);

	input_data_and_cartridge_bus = (u32)GPIO_ReadInputData(GPIOD);
	input_parallel_and_voltage_bus = (u32)GPIO_ReadInputData(GPIOC);
	input_audio = (u32)ADC_GetConversionValue(ADC1);
	//printf("PORT_Get %04x:%04x:%04x\r\n", input_data_and_cartridge_bus, input_parallel_and_voltage_bus, input_audio);

	/*GPIOD_InitStructure.GPIO_Pin = 0
			| GPIO_Pin_0			//D0 (IN floating)
			| GPIO_Pin_1			//D1 (IN floating)
			| GPIO_Pin_2			//D2 (IN floating)
			| GPIO_Pin_3			//D3 (IN floating)
			| GPIO_Pin_4			//D4 (IN floating)
			| GPIO_Pin_5			//D5 (IN floating)
			| GPIO_Pin_6			//D6 (IN floating)
			| GPIO_Pin_7			//D7 (IN floating)
			;*/
	GPIOD_InitStructure.GPIO_Mode = GPIO_Mode_IN_FLOATING;
	GPIO_Init(GPIOD, &GPIOD_InitStructure);

	signals = (((input_data_and_cartridge_bus >> 0) & 0x00ff) << 0)				/* PD0-PD7: D0-D7*/
			| (((input_data_and_cartridge_bus >> RD4) & 0x0001) << 8)			/* PD15: RD4 */
			| (((input_data_and_cartridge_bus >> RD5) & 0x0001) << 9)			/* PD13: RD5 */
			| (((input_parallel_and_voltage_bus >> _IRQ) & 0x0001) << 10)		/* PC8: _IRQ */
			| (((input_parallel_and_voltage_bus >> _EXTSEL) & 0x0001) << 11)	/* PC9: _EXTSEL */
			| (((input_parallel_and_voltage_bus >> _MPD) & 0x0001) << 12)		/* PC10: _MPD */
			| (((input_parallel_and_voltage_bus >> _CASINH) & 0x0001) << 13)	/* PC4: _CASINH */
			| (((input_parallel_and_voltage_bus >> _RESET) & 0x0001) << 14)		/* PC7: _RESET */
			| (((input_parallel_and_voltage_bus >> _HALT) & 0x0001) << 15)		/* PC6: _HALT */
			| (((input_audio >> 0) & 0xffff) << 16)								/* ADC4-ADC15: AUDIO */
			;
	return signals;
}

u32 PORT_Write(u32 signals, u8 data)
{
	u16 address_bus;
	u16 output_data_and_cartridge_bus;
	u16 output_parallel_and_voltage_bus;
	u32 input_data_and_cartridge_bus;
	u32 input_parallel_and_voltage_bus;
	u32 input_audio;

	address_bus = (signals >> 0) & 0xffff;						/* PE0-PE15: A0-A15 */
	output_data_and_cartridge_bus = 0
			| (data << 0)										/* PD0-PD7: D0-D7 */
			| (/*((signals >> 16) & 0x0001)*/0 << PHI2)			/* PD8: phi2 */
			| (/*((signals >> 17) & 0x0001)*/0 << RW)			/* PD9: R/W */
			| (((signals >> 19) & 0x0001) << _CCTL)				/* PD10: /CCTL */
			| (((signals >> 26) & 0x0001) << _RDY)				/* PD11: /RDY */
			| (((signals >> 21) & 0x0001) << _S5)				/* PD12: /S5 */
			//| (1 << RD5)										/* PD13: RD5 */
			| (((signals >> 20) & 0x0001) << _S4)				/* PD14: /S4 */
			//| (1 << RD4)										/* PD15: RD4 */
			;
	output_parallel_and_voltage_bus = 0
			//| (1 << AUDIO)									/* PC0: AUDIO */
			| voltage											/* PC1: /Vpp */
			| (((signals >> 27) & 0x0001) << _CAS)				/* PC2: /CAS */
			| (((signals >> 28) & 0x0001) << _RAS)				/* PC3: /RAS */
			//| (1 << _CASINH)									/* PC4: _CASINH */
			//| GPIO_Pin_5										/* PC5: */
			| (((signals >> 31) & 0x0001) << _HALT)				/* PC6: /HALT */
			| (((signals >> 30) & 0x0001) << _RESET)			/* PC7: /RESET */
			//| (1 << _IRQ)										/* PC8: _IRQ */
			//| (1 << _EXTSEL)									/* PC9: _EXTSEL */
			//| (1 << _MPD)										/* PC10: _MPD */
			| (((signals >> 22) & 0x0001) << _REF)				/* PC11: /REF */
			| (((signals >> 23) & 0x0001) << _D1XX)				/* PC12: /D1XX */
			//| GPIO_Pin_13										/* PC13: */
			//| GPIO_Pin_14										/* PC14: */
			//| GPIO_Pin_15										/* PC15: */
			;

	GPIO_Write(GPIOE, address_bus);
	GPIO_Write(GPIOD, output_data_and_cartridge_bus);
	GPIO_Write(GPIOC, output_parallel_and_voltage_bus);
	//printf("PORT put %04x:%04x:%04x\r\n", address_bus, output_data_and_cartridge_bus, output_parallel_and_voltage_bus);

	/*GPIO_InitStructure.GPIO_Pin = 0
			| GPIO_Pin_0			//D0 (OUT push/pull)
			| GPIO_Pin_1			//D1 (OUT push/pull)
			| GPIO_Pin_2			//D2 (OUT push/pull)
			| GPIO_Pin_3			//D3 (OUT push/pull)
			| GPIO_Pin_4			//D4 (OUT push/pull)
			| GPIO_Pin_5			//D5 (OUT push/pull)
			| GPIO_Pin_6			//D6 (OUT push/pull)
			| GPIO_Pin_7			//D7 (OUT push/pull)
			;*/
	GPIOD_InitStructure.GPIO_Mode = GPIO_Mode_Out_PP;
	/*GPIOD_InitStructure.GPIO_Speed = GPIO_Speed_2MHz;*/
	GPIO_Init(GPIOD, &GPIOD_InitStructure);

#ifdef PHI2_DELAY
	/*Delay_Ns(225);*/
#endif

	GPIO_WriteBit(GPIOD, 1 << PHI2, Bit_SET);

#ifdef PHI2_DELAY
	Delay_Ns(180);
#endif

	GPIO_WriteBit(GPIOD, 1 << PHI2, Bit_RESET);

	input_data_and_cartridge_bus = (u32)GPIO_ReadInputData(GPIOD);
	input_parallel_and_voltage_bus = (u32)GPIO_ReadInputData(GPIOC);
	input_audio = (u32)ADC_GetConversionValue(ADC1);
	//printf("PORT get %04x:%04x:%04x\r\n", input_data_and_cartridge_bus, input_parallel_and_voltage_bus, input_audio);

	/*GPIOD_InitStructure.GPIO_Pin = 0
			| GPIO_Pin_0			//D0 (IN floating)
			| GPIO_Pin_1			//D1 (IN floating)
			| GPIO_Pin_2			//D2 (IN floating)
			| GPIO_Pin_3			//D3 (IN floating)
			| GPIO_Pin_4			//D4 (IN floating)
			| GPIO_Pin_5			//D5 (IN floating)
			| GPIO_Pin_6			//D6 (IN floating)
			| GPIO_Pin_7			//D7 (IN floating)
			;*/
	GPIOD_InitStructure.GPIO_Mode = GPIO_Mode_IN_FLOATING;
	GPIO_Init(GPIOD, &GPIOD_InitStructure);

	signals = 0
			| (((input_data_and_cartridge_bus >> RD4) & 0x0001) << 0)			/* PD15: RD4 */
			| (((input_data_and_cartridge_bus >> RD5) & 0x0001) << 1)			/* PD13: RD5 */
			| (((input_parallel_and_voltage_bus >> _IRQ) & 0x0001) << 2)		/* PC8: _IRQ */
			| (((input_parallel_and_voltage_bus >> _EXTSEL) & 0x0001) << 3)		/* PC9: _EXTSEL */
			| (((input_parallel_and_voltage_bus >> _MPD) & 0x0001) << 4)		/* PC10: _MPD */
			| (((input_parallel_and_voltage_bus >> _CASINH) & 0x0001) << 5)		/* PC4: _CASINH */
			| (((input_parallel_and_voltage_bus >> _RESET) & 0x0001) << 6)		/* PC7: _RESET */
			| (((input_parallel_and_voltage_bus >> _HALT) & 0x0001) << 7)		/* PC6: _HALT */
			| (((input_audio >> 0) & 0xffff) << 8)								/* ADC4-ADC15: AUDIO */
			;
	return signals;
}

bool_t SERIAL_GetByte(u8 *data)
{
	uint32_t i;

	/* set serial communication timeout */

	SysTick->SR &= ~(1 << 0);

	i = SystemCoreClock / 8000000;
	i = ((i * SERIAL_TIMEOUT_NS) + 999) / 1000;	/* ceil */

	SysTick->CMP = i;
	SysTick->CTLR |= (1 << 4);
	SysTick->CTLR |= (1 << 5) | (1 << 0);

	while (USART_GetFlagStatus(USART1, USART_FLAG_RXNE) == RESET)
	{
		if ((SysTick->SR & (1 << 0)) == (1 << 0))
		{
			/* timeout occurred */
			SysTick->CTLR &= ~(1 << 0);
			return FALSE;
		}
		if ( (USART_GetFlagStatus(USART1, USART_FLAG_NE) == SET)
				|| (USART_GetFlagStatus(USART1, USART_FLAG_FE) == SET) )
		{
			/* noise or framing error occurred */
			SysTick->CTLR &= ~(1 << 0);
			USART_ReceiveData(USART1);
			return FALSE;
		}
	}
	SysTick->CTLR &= ~(1 << 0);

	if (USART_GetFlagStatus(USART1, USART_FLAG_ORE) == SET)
	{
		/* overrun error occurred */
		USART_ClearFlag(USART1, USART_FLAG_RXNE);
		return FALSE;
	}

	*data = (u8)USART_ReceiveData(USART1);

	USART_ClearFlag(USART1, USART_FLAG_RXNE);
	return TRUE;
}

bool_t SERIAL_PutByte(u8 data)
{
	uint32_t i;

	/* set serial communication timeout */

	SysTick->SR &= ~(1 << 0);

	i = SystemCoreClock / 8000000;
	i = ((i * SERIAL_TIMEOUT_NS) + 999) / 1000;	/* ceil */

	SysTick->CMP = i;
	SysTick->CTLR |= (1 << 4);
	SysTick->CTLR |= (1 << 5) | (1 << 0);

	while (USART_GetFlagStatus(USART1, USART_FLAG_TXE) == RESET)
	{
		if ((SysTick->SR & (1 << 0)) == (1 << 0))
		{
			/* timeout occurred */
			SysTick->CTLR &= ~(1 << 0);
			return FALSE;
		}
	}
	SysTick->CTLR &= ~(1 << 0);

	USART_SendData(USART1, (u16)data);

	return TRUE;
}

bool_t FRAME_ReadIn(u8 *buffer, u16 count)
{
	u16 i;
	u16 sum;
	u8 crc;

	sum = 0;
	for (i = 0; i < count; i++)
	{
		if (!SERIAL_GetByte(buffer))
		{
			return FALSE;
		}
		sum += *buffer++;
	}
	if (!SERIAL_GetByte(&crc))
	{
		return FALSE;
	}

	return ((sum + (sum >> 8)) & 0xff) == crc;
}

bool_t FRAME_WriteOut(u8 *buffer, u16 count)
{
	u16 i;
	u16 sum;
	u8 crc;

	sum = 0;
	for (i = 0; i < count; i++)
	{
		if (!SERIAL_PutByte(*buffer))
		{
			return FALSE;
		}
		sum += (u16)*buffer++;
	}
	crc = sum + (sum >> 8);
	return SERIAL_PutByte(crc);
}

u32 CART_build_request_read(u16 address)
{
	u8 _cctl;
	u8 _s4;
	u8 _s5;
	_cctl = ((address & 0xff00) == 0xd500) ? 0 : 1;
	_s4 = ((address & 0xe000) == 0x8000) ? 0 : 1;
	_s5 = ((address & 0xe000) == 0xa000) ? 0 : 1;
	return 0
			| (address << 0)	/* address bus */
			//| (0 << 16)		/* PHI2 */
			//| (0 << 17)		/* R/W */
			//| (0 << 18)		/* R/W LATE */
			| (_cctl << 19)		/* _CCTL */
			| (_s4 << 20)		/* _S4 */
			| (_s5 << 21)		/* _S5 */
			| (1 << 22)			/* _REF */
			| (1 << 23)			/* _D1XX */
			//| (0 << 24)		/* ? */
			//| (0 << 25)		/* ? */
			| (1 << 26)			/* _RDY */
			| (1 << 27)			/* _CAS */
			| (1 << 28)			/* _RAS */
			//| (0 << 29)		/* _LATCHRW */
			| (1 << 30)			/* _RESET */
			| (1 << 31)			/* _HALT */
			;
}

u32 CART_build_request_write(u16 address)
{
	u8 _cctl;
	u8 _s4;
	u8 _s5;
	_cctl = ((address & 0xff00) == 0xd500) ? 0 : 1;
	_s4 = ((address & 0xe000) == 0x8000) ? 0 : 1;
	_s5 = ((address & 0xe000) == 0xa000) ? 0 : 1;
	return 0
			| (address << 0)	/* address bus */
			//| (0 << 16)		/* PHI2 */
			//| (0 << 17)		/* R/W */
			//| (0 << 18)		/* R/W LATE */
			| (_cctl << 19)		/* _CCTL */
			| (_s4 << 20)		/* _S4 */
			| (_s5 << 21)		/* _S5 */
			| (1 << 22)			/* _REF */
			| (1 << 23)			/* _D1XX */
			//| (0 << 24)		/* ? */
			//| (0 << 25)		/* ? */
			| (1 << 26)			/* _RDY */
			| (1 << 27)			/* _CAS */
			| (1 << 28)			/* _RAS */
			//| (0 << 29)		/* _LATCHRW */
			| (1 << 30)			/* _RESET */
			| (1 << 31)			/* _HALT */
			;
}

bool_t CART_byte_read(u16 address, u8 *data)
{
	u32 signals;

	signals = CART_build_request_read(address);	/* address bus and cartridge control bus */
	//printf("READ Put %08x\r\n", signals);
	signals = PORT_Read(signals);
	//printf("READ Get %08x\r\n", signals);
	*data = (signals >> 0) & 0xff;			/* data bus */
	return (TRUE
		&& ((signals & (1 << 8)) == 0)		/* RD4 */
		&& ((signals & (1 << 9)) != 0)		/* RD5 */
		);
}

bool_t CART_byte_write(u16 address, u8 data)
{
	u32 signals;

	signals = CART_build_request_write(address);	/* address bus and cartridge control bus */
	//printf("WRITE Put %08x\r\n", signals);
	signals = PORT_Write(signals, data);
	//printf("WRITE Get %08x\r\n", signals);
	return (TRUE
		&& ((signals & (1 << 0)) == 0)		/* RD4 */
		&& ((signals & (1 << 1)) != 0)		/* RD5 */
		);
}

bool_t CART_program_sequence(u32 address, u8 data)
{
	u32 chip;

	chip = address & 0x1f8000;

	if (!(CART_byte_write(0xd500 | (((chip | 0x5555) & 0x1fe000) >> 13), 0xff)))
	{
		return FALSE;
	}
	if (!(CART_byte_write(0xa000 | (((chip | 0x5555) & 0x001fff) >> 0), 0xaa)))
	{
		return FALSE;
	}

	if (!(CART_byte_write(0xd500 | (((chip | 0x2aaa) & 0x1fe000) >> 13), 0xff)))
	{
		return FALSE;
	}
	if (!(CART_byte_write(0xa000 | (((chip | 0x2aaa) & 0x001fff) >> 0), 0x55)))
	{
		return FALSE;
	}

	if (!(CART_byte_write(0xd500 | ((address & 0x1fe000) >> 13), 0xff)))
	{
		return FALSE;
	}
	if (!(CART_byte_write(0xa000 | ((address & 0x001fff) >> 0), data)))
	{
		return FALSE;
	}

	return TRUE;
}

bool_t CART_byte_program_sequence(u32 address, u8 data)
{
	u32 chip;

	chip = address & 0x1f8000;

	if (!(CART_program_sequence(chip | 0x5555, 0xa0)))
	{
		return FALSE;
	}

	if (!(CART_byte_write(0xd500 | ((address & 0x1fe000) >> 13), 0xff)))
	{
		return FALSE;
	}
	if (!(CART_byte_write(0xa000 | ((address & 0x001fff) >> 0), data)))
	{
		return FALSE;
	}

	Delay_Us(20);

	return TRUE;
}

bool_t CART_sector_erase_sequence(u32 address)
{
	u32 chip;

	chip = address & 0x1f8000;

	if (!(CART_program_sequence(chip | 0x5555, 0x80)))
	{
		return FALSE;
	}

	if (!(CART_program_sequence(address, 0x30)))
	{
		return FALSE;
	}

	Delay_Ms(100);

	return TRUE;
}

bool_t CART_chip_erase_sequence(u32 address)
{
	u32 chip;

	chip = address & 0x1f8000;

	if (!(CART_program_sequence(chip | 0x5555, 0x80)))
	{
		return FALSE;
	}

	if (!(CART_program_sequence(chip | 0x5555, 0x10)))
	{
		return FALSE;
	}

	Delay_Ms(100);

	return TRUE;
}

bool_t CART_software_id_entry_sequence(u32 address)
{
	u32 chip;

	chip = address & 0x1f8000;

	if (!(CART_program_sequence(chip | 0x5555, 0x90)))
	{
		return FALSE;
	}

	Delay_Ns(150);

	return TRUE;
}

#ifdef LONG_SOFTWARE_ID_EXIT

bool_t CART_software_id_exit_sequence(u32 address)
{
	u32 chip;

	chip = address & 0x1f8000;

	if (!(CART_program_sequence(chip | 0x5555, 0xf0)))
	{
		return FALSE;
	}

	Delay_Ns(150);

	return TRUE;
}

#else /*LONG_SOFTWARE_ID_EXIT*/

bool_t CART_software_id_exit_sequence(u32 address)
{
	u32 chip;

	chip = address & 0x1f8000;

	if (!(CART_byte_write(0xd500 | (((chip | 0x00000) & 0x1fe000) >> 13), 0xff)))
	{
		return FALSE;
	}
	if (!(CART_byte_write(0xa000 | (((chip | 0x00000) & 0x001fff) >> 0), 0xf0)))
	{
		return FALSE;
	}

	Delay_Ns(150);

	return TRUE;
}

#endif /*LONG_SOFTWARE_ID_EXIT*/


/*
 * Manufacturer:
 * 0xad: Hyundai Electronics
 * Device:
 * 0xa4: HY29F040A - 64K sector
 *
 * Manufacturer:
 * 0xbf: SST
 * Device:
 * 0xb5: SST39SF010A - 4K sector
 * 0xb6: SST39SF020A - 4K sector
 * 0xb7: SST39SF040 - 4K sector
 */
static bool_t CART_read_cartridge_memory_chip_identifier_internal(u32 address, u8 *manufacturer_id, u8 *device_id)
{
	u32 chip;

	chip = address & 0x1f8000;

	if (!(CART_byte_write(0xd500 | (((chip | 0) & 0x1fe000) >> 13), 0xff)))
	{
		return FALSE;
	}
	if (!(CART_byte_read(0xa000 | (((chip | 0) & 0x001fff) >> 0), manufacturer_id)))
	{
		return FALSE;
	}

	if (!(CART_byte_write(0xd500 | (((chip | 1) & 0x1fe000) >> 13), 0xff)))
	{
		return FALSE;
	}
	if (!(CART_byte_read(0xa000 | (((chip | 1) & 0x001fff) >> 0), device_id)))
	{
		return FALSE;
	}

	return TRUE;
}

bool_t CART_read_cartridge_memory_chip_identifier(u32 address, u8 *manufacturer_id, u8 *device_id)
{
	bool_t status = TRUE;

	if (CART_software_id_entry_sequence(address))
	{
		status = CART_read_cartridge_memory_chip_identifier_internal(address, manufacturer_id, device_id);
	}

	if (!(CART_software_id_exit_sequence(address)))
	{
		return FALSE;
	}

	return status;
}


void VRHI_InfoCommand(void)
{
	FRAME_WriteOut(INFO, sizeof(INFO));
}

void VRHI_PowerCommand(void)
{
	u8 power;

	if (FRAME_ReadIn(buffer, 1))
	{
		power = 0
				| (buffer[0] << 0)
				;

		//printf("POWER Put %02x\r\n", power);
	    power = PORT_Power(power);
		//printf("POWER Get %02x\r\n", power);

		buffer[0] = (power >> 0);
		FRAME_WriteOut(buffer, 1);
	}
}

void VRHI_SignalsCommand(void)
{
	u32 signals;
	u8 data;

	if (FRAME_ReadIn(buffer, 5))
	{
		signals = 0
				| (buffer[0] << 0)				/* address bus lsb */
				| (buffer[1] << 8)				/* address bus msb */
				| (buffer[2] << 16)				/* cartridge and parallel bus */
				| (buffer[3] << 24)				/* parallel bus */
				;
		data = buffer[4];						/* data bus */

		//printf("SIGNAL Put %08x\r\n", signals);
		signals = PORT_Signals(signals, data);
		//printf("SIGNAL Get %08x\r\n", signals);

		buffer[0] = (signals >> 0);				/* data bus */
		buffer[1] = (signals >> 8);				/* cartridge and parallel bus */
		buffer[2] = (signals >> 16);			/* audio lsb */
		buffer[3] = (signals >> 24);			/* audio msb */
		FRAME_WriteOut(buffer, 4);
	}
}

void VRHI_ReadCommand(void)
{
	u32 signals;

	if (FRAME_ReadIn(buffer, 4))
	{
		signals = 0
				| (buffer[0] << 0)				/* address bus lsb */
				| (buffer[1] << 8)				/* address bus msb */
				| (buffer[2] << 16)				/* cartridge and parallel bus */
				| (buffer[3] << 24)				/* parallel bus */
				;

		//printf("READ Put %08x\r\n", signals);
		signals = PORT_Read(signals);
		//printf("READ Get %08x\r\n", signals);

		buffer[0] = (signals >> 0);				/* data bus */
		buffer[1] = (signals >> 8);				/* cartridge and parallel bus */
		buffer[2] = (signals >> 16);			/* audio lsb */
		buffer[3] = (signals >> 24);			/* audio msb */
		FRAME_WriteOut(buffer, 4);
	}
}

void VRHI_WriteCommand(void)
{
	u32 signals;
	u8 data;

	if (FRAME_ReadIn(buffer, 5))
	{
		signals = 0
				| (buffer[0] << 0)				/* address bus lsb */
				| (buffer[1] << 8)				/* address bus msb */
				| (buffer[2] << 16)				/* cartridge and parallel bus */
				| (buffer[3] << 24)				/* parallel bus */
				;
		data = buffer[4];						/* data bus */

		//printf("WRITE Put %08x\r\n", signals);
		signals = PORT_Write(signals, data);
		//printf("WRITE Get %08x\r\n", signals);

		buffer[0] = (signals >> 0);				/* cartridge and parallel bus */
		buffer[1] = (signals >> 8);				/* audio lsb */
		buffer[2] = (signals >> 16);			/* audio msb */
		FRAME_WriteOut(buffer, 3);
	}
}

void VRHI_ResetCommand(void)
{
	PORT_Reset();
	SERIAL_PutByte(0);
}

void VRHI_ChipInfoCommand(void)
{
	u32 address;

	if (FRAME_ReadIn(buffer, 3))
	{
		address = buffer[0] | (buffer[1] << 8) | (buffer[2] << 16);

		if (!(CART_read_cartridge_memory_chip_identifier(address, buffer+0, buffer+1)))
		{
			return;
		}

		FRAME_WriteOut(buffer, 2);
	}
}

void VRHI_ChipEraseCommand(void)
{
	u32 address;

	if (FRAME_ReadIn(buffer, 3))
	{
		address = buffer[0] | (buffer[1] << 8) | (buffer[2] << 16);

		if (!(CART_chip_erase_sequence(address)))
		{
			return;
		}

		SERIAL_PutByte(0);
	}
}

void VRHI_SectorEraseCommand(void)
{
	u32 address;

	if (FRAME_ReadIn(buffer, 3))
	{
		address = buffer[0] | (buffer[1] << 8) | (buffer[2] << 16);

		if (!(CART_sector_erase_sequence(address)))
		{
			return;
		}

		SERIAL_PutByte(0);
	}
}

void VRHI_BankReadCommand(void)
{
	u16 i;
	u8 bank;

	if (FRAME_ReadIn(buffer, 1))
	{
		bank = buffer[0];

		if (!(CART_byte_write(0xd500 | bank, 0xff)))
		{
			return;
		}

		for (i = 0; i < BANK_SIZE; i++)
		{
			if (!(CART_byte_read(0xa000 | i, buffer+i)))
			{
				return;
			}
		}

		FRAME_WriteOut(buffer, BANK_SIZE);
	}
}

void VRHI_BankWriteCommand(void)
{
	u16 i;
	u8 bank;
	u32 address;

	if (FRAME_ReadIn(buffer, 1+BANK_SIZE))
	{
		bank = buffer[0];

		address = (bank << 13);
		for (i = 0; i < BANK_SIZE; i++)
		{
			if (!(CART_byte_program_sequence(address | i, buffer[1+i])))
			{
				return;
			}
		}

		SERIAL_PutByte(0);
	}
}

/*********************************************************************
 * @fn      main
 *
 * @brief   Main program.
 *
 * @return  none
 */
int main(void)
{
	u8 command;

	NVIC_PriorityGroupConfig(NVIC_PriorityGroup_2);
	Delay_Init();
	//USART_Printf_Init(115200);

    UART_Prepare();
    GPIO_Prepare();
    PORT_Reset();

	//printf("VRHI %d.%d\r\n", FW_VERSION, FW_REVISION);

	while (TRUE)
    {
		if (SERIAL_GetByte(&command))
		{
			switch (command)
			{
			case 'i':
				VRHI_ChipInfoCommand();
				break;
			case 'c':
				VRHI_ChipEraseCommand();
				break;
			case 'e':
				VRHI_SectorEraseCommand();
				break;
			case 'r':
				VRHI_BankReadCommand();
				break;
			case 'w':
				VRHI_BankWriteCommand();
				break;

			case '?':
				VRHI_InfoCommand();
				break;
			case 'P':
				VRHI_PowerCommand();
				break;
			case 'R':
				VRHI_ReadCommand();
				break;
			case 'S':
				VRHI_SignalsCommand();
				break;
			case 'W':
				VRHI_WriteCommand();
				break;
			default:
				VRHI_ResetCommand();
			}
		}
	}
}
