
/*
 * VRHI cartridge programator
 * supported devices:
 * 1. jAtariCart (c) by Jakub Husak
 * 1MB RAM two chips HY29F040A with 64KB sector (A16..)
 * 2. MaxFlash128KBPlus or GalCart (c) by Pancio
 * 512KB RAM chip SST39SF040 with 4KB sector (A12..)
 * TODO:
 * 1. verify byte programming
 * 2. overwrite already programmed byte
 * 3. erase sector when error
 * 4. break programming when two erases failed
 * 5. check byte programmed when timeout error
 * 6. repeat read byte when checksum error
 * 7. repeat programming when wrong byte
 * 8. check type of cartridge and memory size and chips type and sector size
 * 9. DQ7 data polling instead of time waiting
 * 10. cartridge bank (4/8/16K) read/program commands
 */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include <vrhi.h>


/*#define LONG_SOFTWARE_ID_EXIT*/

/* Number of banks. */
#define BANKCOUNT 0x40

/* Bank size. */
#define BANKSIZE 0x2000

/* Default device. */
#define DEFAULT_DEVICE "/dev/ttyUSB0"

enum
{
	FALSE = (0 == 1),
	TRUE = (1 == 1)
};

/* Cartridge memory buffer. */
static char buffer[ BANKSIZE * BANKCOUNT ];


vrhi_status_t initialize_communication(void)
{
	vrhi_status_t status;
	vrhi_response_info_t response_info;

	printf("Reset device\n");
	status = vrhi_reset();
	if (status != EOK)
	{
		return status;
	}

	printf("Reading device information\n");
	status = vrhi_info(&response_info);
	if (status != EOK)
	{
		return status;
	}
	printf("Signature: %.4s\n", response_info.signature);
	printf("Platform: %.2s\n", response_info.platform);
	printf("Firmware: %d.%d\n", response_info.version, response_info.revision);

	return EOK;
}

void build_request_read(vrhi_request_read_t *request, unsigned short address)
{
	request->address = address & 0xffff;

	/*
	request->phi2 = HIGH;
	request->rw = HIGH;
	request->rwlate = HIGH;
	*/
	request->_cctl = !((address & 0xff00) == 0xd500);
	request->_s4 = !((address & 0xe000) == 0x8000);
	request->_s5 = !((address & 0xe000) == 0xa000);

	request->_ref = HIGH;
	request->_d1xx = HIGH;
	request->_rdy = HIGH;
	request->_cas = HIGH;
	request->_ras = HIGH;
	/*
	request->_latchrw = HIGH;
	*/
	request->_reset = HIGH;
	request->_halt = HIGH;
}

void build_request_write(vrhi_request_write_t *request, unsigned short address, unsigned char data)
{
	request->address = address & 0xffff;

	/*
	request->phi2 = HIGH;
	request->rw = HIGH;
	request->rwlate = HIGH;
	*/
	request->_s4 = !((address & 0xe000) == 0x8000);
	request->_s5 = !((address & 0xe000) == 0xa000);
	request->_cctl = !((address & 0xff00) == 0xd500);

	request->data = data;

	request->_ref = HIGH;
	request->_d1xx = HIGH;
	request->_rdy = HIGH;
	request->_cas = HIGH;
	request->_ras = HIGH;
	/*
	request->_latchrw = HIGH;
	*/
	request->_reset = HIGH;
	request->_halt = HIGH;
}

vrhi_status_t cartridge_power_on(void)
{
	vrhi_status_t status;
	vrhi_request_power_t request_power = {
		._vpp = LOW,	/* power on */
	};
	vrhi_response_power_t response_power;

	printf("Cartridge power on\n");

	status = vrhi_power(&request_power, &response_power);
	if (status != EOK)
	{
		return status;
	}
	if (usleep(20) != 0)
	{
		return EERRNO;
	}

	return status;
}

vrhi_status_t cartridge_power_off(void)
{
	vrhi_status_t status;
	vrhi_request_power_t request_power = {
		._vpp = HIGH,	/* power off */
	};
	vrhi_response_power_t response_power;

	printf("Cartridge power off\n");

	status = vrhi_power(&request_power, &response_power);
	if (status != EOK)
	{
		return status;
	}
	if (usleep(20) != 0)
	{
		return EERRNO;
	}

	return status;
}

vrhi_status_t cartridge_read(vrhi_request_read_t *request_read, vrhi_response_read_t *response_read)
{
	vrhi_status_t status;

	status = vrhi_read(request_read, response_read);
	if (status != EOK)
	{
		return status;
	}
	if (!((response_read->rd4 == LOW) && (response_read->rd5 == HIGH)))
	{
		printf("Wrong signals state: RD4=%d (0) and RD5=%d (1) on READ\n", response_read->rd4, response_read->rd5);
		return status;
	}
	return EOK;
}

vrhi_status_t cartridge_write(vrhi_request_write_t *request_write, vrhi_response_write_t *response_write)
{
	vrhi_status_t status;

	status = vrhi_write(request_write, response_write);
	if (status != EOK)
	{
		return status;
	}
	if (!((response_write->rd4 == LOW) && (response_write->rd5 == HIGH)))
	{
		printf("Wrong signals state: RD4=%d (0) and RD5=%d (1) on WRITE\n", response_write->rd4, response_write->rd5);
		return status;
	}
	return EOK;
}

/*
 * Sequence of triple bytes used to enter or exit a special mode.
 * 1. WRITE $AA -> $5555
 * 2. WRITE $55 -> $2AAA
 * 3. WRITE data -> address
 * @param address address in 3rd step
 * @param data data in 3rd step
 * @returns operation status
 */
vrhi_status_t program_sequence(unsigned int address, unsigned char data)
{
	vrhi_status_t status;
	vrhi_request_write_t request_write;
	vrhi_response_write_t response_write;

	/* 1. WRITE $AA -> $5555 */
	build_request_write(&request_write, 0xd500 | ((0x5555 & 0xe000) >> 13), 0xff);
	status = cartridge_write(&request_write, &response_write);
	if (status != EOK)
	{
		return status;
	}
	build_request_write(&request_write, 0xa000 | ((0x5555 & 0x1fff) >> 0), 0xaa);
	status = cartridge_write(&request_write, &response_write);
	if (status != EOK)
	{
		return status;
	}

	/* 2. WRITE $55 -> $2AAA */
	build_request_write(&request_write, 0xd500 | ((0x2aaa & 0xe000) >> 13), 0xff);
	status = cartridge_write(&request_write, &response_write);
	if (status != EOK)
	{
		return status;
	}
	build_request_write(&request_write, 0xa000 | ((0x2aaa & 0x1fff) >> 0), 0x55);
	status = cartridge_write(&request_write, &response_write);
	if (status != EOK)
	{
		return status;
	}

	/* 3. WRITE data -> address */
	build_request_write(&request_write, 0xd500 | ((address & 0x7e000) >> 13), 0xff);
	status = cartridge_write(&request_write, &response_write);
	if (status != EOK)
	{
		return status;
	}
	build_request_write(&request_write, 0xa000 | ((address & 0x01fff) >> 0), data);
	status = cartridge_write(&request_write, &response_write);
	if (status != EOK)
	{
		return status;
	}

	return EOK;
}

/*
 * Byte-Program sequence.
 * 1. WRITE $AA -> $5555
 * 2. WRITE $55 -> $2AAA
 * 3. WRITE $A0 -> $5555
 * 4. WRITE data -> byte address
 * 5. wait TBP = 20 microseconds
 * @returns operation status
 */
vrhi_status_t byte_program_sequence(unsigned char bank, unsigned short address, unsigned char data)
{
	vrhi_status_t status;
	vrhi_request_write_t request_write;
	vrhi_response_write_t response_write;

	/*
	 * 1. WRITE $AA -> $5555
	 * 2. WRITE $55 -> $2AAA
	 * 3. WRITE $A0 -> $5555
	 */
	status = program_sequence(0x5555, 0xa0);
	if (status != EOK)
	{
		return status;
	}
	/* 4. WRITE data -> byte address */
	build_request_write(&request_write, 0xd500 | bank, 0xff);
	status = cartridge_write(&request_write, &response_write);
	if (status != EOK)
	{
		return status;
	}
	build_request_write(&request_write, 0xa000 | address, data);
	status = cartridge_write(&request_write, &response_write);
	if (status != EOK)
	{
		return status;
	}

	/* 5. wait TBP = max 20 microseconds */
	if (usleep(20) != 0)
	{
		return EERRNO;
	}

	return EOK;
}

/*
 * Sector-Erase sequence.
 * 1. WRITE $AA -> $5555
 * 2. WRITE $55 -> $2AAA
 * 3. WRITE $80 -> $5555
 * 4. WRITE $AA -> $5555
 * 5. WRITE $55 -> $2AAA
 * 6. WRITE $30 -> sector address
 * 7. wait TSCE = max 100 miliseconds
 * WARNING: Sector size is 4KB and is different than cartridge bank size 8KB.
 * @param bank 8KB bank number
 * @param address address in bank ($0000-$1FFF)
 * @returns operation status
 */
vrhi_status_t sector_erase_sequence(unsigned char bank, unsigned short address)
{
	vrhi_status_t status;

	/*
	 * 1. WRITE $AA -> $5555
	 * 2. WRITE $55 -> $2AAA
	 * 3. WRITE $80 -> $5555
	 */
	status = program_sequence(0x5555, 0x80);
	if (status != EOK)
	{
		return status;
	}
	/*
	 * 4. WRITE $AA -> $5555
	 * 5. WRITE $55 -> $2AAA
	 * 6. WRITE $30 -> sector address
	 */
	status = program_sequence(((unsigned int)bank << 13) | address, 0x30);
	if (status != EOK)
	{
		return status;
	}

	/* 7. wait TSCE = max 100 miliseconds */
	if (usleep(100000) != 0)
	{
		return EERRNO;
	}

	return EOK;
}

/*
 * Chip-Erase sequence.
 * 1. WRITE $AA -> $5555
 * 2. WRITE $55 -> $2AAA
 * 3. WRITE $80 -> $5555
 * 4. WRITE $AA -> $5555
 * 5. WRITE $55 -> $2AAA
 * 6. WRITE $10 -> $5555
 * 7. wait TSCE = max 100 miliseconds
 * @returns operation status
 */
vrhi_status_t chip_erase_sequence(void)
{
	vrhi_status_t status;

	/*
	 * 1. WRITE $AA -> $5555
	 * 2. WRITE $55 -> $2AAA
	 * 3. WRITE $80 -> $5555
	 */
	status = program_sequence(0x5555, 0x80);
	if (status != EOK)
	{
		return status;
	}
	/*
	 * 4. WRITE $AA -> $5555
	 * 5. WRITE $55 -> $2AAA
	 * 6. WRITE $10 -> $5555
	 */
	status = program_sequence(0x5555, 0x10);
	if (status != EOK)
	{
		return status;
	}

	/* 7. wait TSCE = max 100 miliseconds */
	if (usleep(100000) != 0)
	{
		return EERRNO;
	}

	return EOK;
}

/*
 * Software-ID-Entry seqeunce.
 * 1. WRITE $AA -> $5555
 * 2. WRITE $55 -> $2AAA
 * 3. WRITE $90 -> $5555
 * 4. wait TIDA = max 150 nanoseconds
 * @returns operation status
 */
vrhi_status_t software_id_entry_sequence(void)
{
	vrhi_status_t status;

	/*
	 * 1. WRITE $AA -> $5555
	 * 2. WRITE $55 -> $2AAA
	 * 3. WRITE $90 -> $5555
	 */
	status = program_sequence(0x5555, 0x90);
	if (status != EOK)
	{
		return status;
	}

	/* 7. wait TIDA = max 150 nanoseconds */
	if (usleep(0.150) != 0)
	{
		return EERRNO;
	}

	return EOK;
}

#ifdef LONG_SOFTWARE_ID_EXIT
/*
 * Software-Id-Exit sequence.
 * 1. WRITE $AA -> $5555
 * 2. WRITE $55 -> $2AAA
 * 3. WRITE $F0 -> $5555
 * 4. wait TIDA = max 150 nanoseconds
 * @returns operation status
 */
vrhi_status_t software_id_exit_sequence(void)
{
	vrhi_status_t status;

	/*
	 * 1. WRITE $AA -> $5555
	 * 2. WRITE $55 -> $2AAA
	 * 3. WRITE $F0 -> $5555
	 */
	status = program_sequence(0x5555, 0xf0);
	if (status != EOK)
	{
		return status;
	}

	/* 4. wait TIDA = max 150 nanoseconds */
	if (usleep(0.150) != 0)
	{
		return EERRNO;
	}

	return EOK;
}
#else
/*
 * Software-Id-Exit sequence.
 * 1. WRITE $F0 -> $xxxx
 * 2. wait TIDA = max 150 nanoseconds
 * @returns operation status
 */
vrhi_status_t software_id_exit_sequence(void)
{
	vrhi_status_t status;
	vrhi_request_write_t request_write;
	vrhi_response_write_t response_write;

	/* 1. WRITE $F0 -> $xxxx */
	build_request_write(&request_write, 0xd500 | 0x0000, 0xff);
	status = cartridge_write(&request_write, &response_write);
	if (status != EOK)
	{
		return status;
	}
	build_request_write(&request_write, 0xa000 | 0x0000, 0xf0);
	status = cartridge_write(&request_write, &response_write);
	if (status != EOK)
	{
		return status;
	}

	/* 2. wait TIDA = max 150 nanoseconds */
	if (usleep(0.150) != 0)
	{
		return EERRNO;
	}

	return EOK;
}
#endif

vrhi_status_t read_cartridge_memory_chip_identifier()
{
	vrhi_status_t status;
	vrhi_request_read_t request_read;
	vrhi_response_read_t response_read;
	vrhi_request_write_t request_write;
	vrhi_response_write_t response_write;
	unsigned char manufacturer_id, device_id;
	char *manufacturer_name, *device_name;

	printf("Reading cartridge memory chip identifier\n");

	status = software_id_entry_sequence();
	if (status != EOK)
	{
		return status;
	}

	/* manufacturer id */

	build_request_write(&request_write, 0xd500 | ((0x00000 & 0x7e000) >> 13), 0xff);
	status = cartridge_write(&request_write, &response_write);
	if (status != EOK)
	{
		return status;
	}

	build_request_read(&request_read, 0xa000 | ((0x00000 & 0x01fff) >> 0));
	status = cartridge_read(&request_read, &response_read);
	if (status != EOK)
	{
		return status;
	}
	manufacturer_id = response_read.data;

	/* device id */

	build_request_write(&request_write, 0xd500 | ((0x00001 & 0x7e000) >> 13), 0xff);
	status = cartridge_write(&request_write, &response_write);
	if (status != EOK)
	{
		return status;
	}

	build_request_read(&request_read, 0xa000 | ((0x00001 & 0x01fff) >> 0));
	status = cartridge_read(&request_read, &response_read);
	if (status != EOK)
	{
		return status;
	}
	device_id = response_read.data;

	status = software_id_exit_sequence();
	if (status != EOK)
	{
		return status;
	}

	switch (manufacturer_id)
	{
	case 0xad:
		manufacturer_name = "Hyundai Electronics";
		break;
	case 0xbf:
		manufacturer_name = "SST";
		break;
	default:
		manufacturer_name = "unknown";
	}

	switch (device_id)
	{
	case 0xa4:
		device_name = "HY29F040A";
		break;
	case 0xb5:
		device_name = "SST39SF010A";
		break;
	case 0xb6:
		device_name = "SST39SF020A";
		break;
	case 0xb7:
		device_name = "SST39SF040";
		break;
	default:
		device_name = "unknown";
	}

	printf("Manufacturer: %s ($%02x)\n", manufacturer_name, manufacturer_id);
	printf("Device: %s ($%02x)\n", device_name, device_id);

	return EOK;
}

vrhi_status_t read_cartridge_bank(unsigned char bank, char *buffer, unsigned short size)
{
	vrhi_status_t status;
	vrhi_request_read_t request_read;
	vrhi_response_read_t response_read;
	vrhi_request_write_t request_write;
	vrhi_response_write_t response_write;
	unsigned short address;

	printf("Reading memory bank $%02x/$%02x content\n", bank, BANKCOUNT);
	build_request_write(&request_write, 0xd500 | bank, 0xff);
	status = cartridge_write(&request_write, &response_write);
	if (status != EOK)
	{
		return status;
	}

	for (address = 0; address < size; address++)
	{
		build_request_read(&request_read, 0xa000 | address);
		status = cartridge_read(&request_read, &response_read);
		if (status != EOK)
		{
			printf("DEBUG: $%02x:%04x\n", bank, address);
			return status;
		}
		buffer[address] = response_read.data;
	}

	return EOK;
}

vrhi_status_t read_cartridge_memory_content(void)
{
	vrhi_status_t status;
	unsigned char bank;

	printf("Reading content of the cartridge\n");
	for (bank = 0; bank < BANKCOUNT; bank++)
	{
		status = read_cartridge_bank(bank, buffer + BANKSIZE * bank, BANKSIZE);
		if (status != EOK)
		{
			printf("Problem in reading bank $%02x\n", bank);
			break;
		}
	}

	return status;
}

vrhi_status_t write_cartridge_bank(unsigned char bank, char *buffer, unsigned short size)
{
	vrhi_status_t status;
	unsigned short address;

	printf("Writing memory bank $%02x/$%02x content\n", bank, BANKCOUNT);
	for (address = 0; address < size; address++)
	{
		status = byte_program_sequence(bank, address, buffer[address]);
		if (status != EOK)
		{
			printf("DEBUG: $%02x:%04x $%02x\n", bank, address, (unsigned char)buffer[address]);
			return status;
		}
	}

	return EOK;
}

vrhi_status_t write_cartridge_memory_content(void)
{
	vrhi_status_t status;
	unsigned char bank;

	printf("Erase chip on the cartridge\n");
	status = chip_erase_sequence();
	if (status != EOK)
	{
		return status;
	}

	printf("Writing content to the cartridge\n");
	for (bank = 0; bank < BANKCOUNT; bank++)
	{
		status = write_cartridge_bank(bank, buffer + BANKSIZE * bank, BANKSIZE);
		if (status != EOK)
		{
			printf("Problem in writing bank $%02x\n", bank);
			break;
		}
	}

	return status;
}

vrhi_status_t read_rom_file(char *filename)
{
	vrhi_status_t status;
	FILE *f;

	printf("Reading file %s\n", filename);
	f = fopen(filename, "rb");
	if (f != NULL)
	{
		ssize_t count, size;

		size = sizeof(buffer);

		memset(buffer, 0xff, size);

		count = fread(buffer, 1, size, f);
		if ((count & 0x1fff) != 0)
		{
			printf("File has too few bytes.\n");
			status = EOK;
		}
		else
		{
			status = EOK;
		}
		if (fclose(f) != 0)
		{
			status = EERRNO;
		}
	}
	else
	{
		status = EERRNO;
	}

	return status;
}

vrhi_status_t write_rom_file(char *filename)
{
	vrhi_status_t status;
	FILE *f;

	printf("Writing file %s\n", filename);
	f = fopen(filename, "wb");
	if (f != NULL)
	{
		ssize_t count, size;

		size = sizeof(buffer);
		count = fwrite(buffer, 1, size, f);
		if (count != size)
		{
			status = ESTOOFEW;
		}
		else
		{
			status = EOK;
		}
		if (fclose(f) != 0)
		{
			status = EERRNO;
		}
	}
	else
	{
		status = EERRNO;
	}

	return status;
}

vrhi_status_t process(char *operation, char *filename)
{
	vrhi_status_t status;

	status = read_cartridge_memory_chip_identifier();
	if (status != EOK)
	{
		return status;
	}

	switch (*operation)
	{
	case 'r'/*ead*/:
		status = read_cartridge_memory_content();
		if (status != EOK)
		{
			return status;
		}
		status = write_rom_file(filename);
		break;
	case 'p'/*rogram*/:
		status = read_rom_file(filename);
		if (status != EOK)
		{
			return status;
		}
		status = write_cartridge_memory_content();
		break;
	default:
		printf("Unexpected operation: %s\n", operation);
	}

	return status;
}

int main(int argc, char *argv[])
{
	vrhi_status_t status;
	int opt, help;
	char *operation, *serial, *filename;

	printf("Virtual to Real Hardware Interface Cartridge Programator\n"
		"API: %d.%d\n"
		"Supported devices: MaxPlus 512K cartridge\n",
		vrhi_version, vrhi_revision);

	operation = "read";
	serial = DEFAULT_DEVICE;
	filename = NULL;
	help = FALSE;
	while ((opt = getopt(argc, argv, "ps:")) >= 0)
	{
		switch (opt)
		{
		case 'p':
			operation = "program";
			break;
		case 's':
			serial = optarg;
			break;
		default:
			help = TRUE;
		}
	}
	if (optind < argc)
	{
		filename = argv[optind];
	}

	if ( help
		|| (filename == NULL)
	)
	{
		printf(
			"\n"
			"Usage: %s [-p] [-s device] image.rom\n"
			"\n"
			"Where:\n"
			"-p\t\tprogram image.rom on the device\n"
			"\t\t(reading from the device into file is the default action)\n"
			"-s device\tthe serial device where programator device is connected to\n"
			"\t\t(" DEFAULT_DEVICE " is taken by default)\n"
			"image.rom\tthe ROM image file\n"
			"\n"
			"Send bug reports and feature requests to Jerzy Kut <mono@protolabo.pl>\n",
			argv[0]
		);
		exit(EXIT_FAILURE);
	}

	printf("Opening %s\n", serial);
	status = vrhi_open(serial);
	if (status == EOK)
	{
		status = initialize_communication();
		if (status == EOK)
		{
			vrhi_status_t status_off;

			status = cartridge_power_on();
			if (status == EOK)
			{
				status = process(operation, filename);
			}
			status_off = cartridge_power_off();
			if (status == EOK)
			{
				status = status_off;
			}
		}
	}
	if (status != EOK)
	{
		char *errstr;

		switch (status)
		{
		case EWPLATFORM:
			errstr = "Wrong device platform";
			break;
		case EWSIGNATURE:
			errstr = "Wrong device signature";
			break;
		case EWCSUM:
			errstr = "Wrong checksum";
			break;
		case ENEXISTS:
			errstr = "Device not exists";
			break;
		case ERTOOFEW:
			errstr = "Too few data received";
			break;
		case ESTOOFEW:
			errstr = "Too few data sent";
			break;
		case ENOPENED:
			errstr = "Device not opened";
			break;
		case EAOPENED:
			errstr = "Device already opened";
			break;
		case EERRNO:
			errstr = strerror(errno);
			status = errno;
			break;
		default:
			errstr = "Unknown error";
		}
		printf("Error %d: %s\n", status, errstr);
	}

	printf("Closing %s\n", serial);
	vrhi_close();

	exit(EXIT_SUCCESS);
}

