
/*
 * VRHI cartridge programator
 * supported devices:
 * 1. jAtariCart (c) by Jakub Husak
 * 1MB RAM two chips HY29F040A with 64KB sector (A16..)
 * 2. MaxFlash128KBPlus or GalCart (c) by Pancio
 * 512KB RAM chip SST39SF040 with 4KB sector (A12..)
 * TODO:
 * 1. check type of cartridge and memory size and chips type and sector size
 * 2. cartridge bank (4K/8K/16K) read/program commands
 */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include <vrhi.h>


/* Number of banks. */
#define BANK_COUNT 0x40

/* Default device. */
#define DEFAULT_DEVICE "/dev/ttyUSB0"

/* Default cartridge type. */
#define CART_NONE 0
#define CART_COUNT 256

/* Number of communication retries. */
#define RETRY_COUNT 10

enum
{
	FALSE = (0 == 1),
	TRUE = (1 == 1)
};

typedef struct
{
	int type;
	char *name;
	int programmable;
	unsigned int chip_count;
	unsigned int sector_size;
	unsigned int sector_count;
}
cartridge_t;

/* Cartridge types. */
static const cartridge_t CARTRIDGES[] =
{
	{   0, NULL,                                        FALSE, 0, 0, 0 },

	{   1, "Standard 8 KB cartridge",                   FALSE, 0, 0, 0 },
	{   2, "Standard 16 KB cartridge",                  FALSE, 0, 0, 0 },
	{   3, "OSS two chip 16 KB cartridge (034M)",       FALSE, 0, 0, 0 },
	{   4, "Standard 32 KB 5200 cartridge",             FALSE, 0, 0, 0 },
	{   5, "DB 32 KB cartridge",                        FALSE, 0, 0, 0 },
	{   6, "Two chip 16 KB 5200 cartridge",             FALSE, 0, 0, 0 },
	{   7, "Bounty Bob 40 KB 5200 cartridge",           FALSE, 0, 0, 0 },
	{   8, "64 KB Williams cartridge",                  FALSE, 0, 0, 0 },
	{   9, "Express 64 KB cartridge",                   FALSE, 0, 0, 0 },

	{  10, "Diamond 64 KB cartridge",                   FALSE, 0, 0, 0 },
	{  11, "SpartaDOS X 64 KB cartridge",               FALSE, 0, 0, 0 },
	{  12, "XEGS 32 KB cartridge",                      FALSE, 0, 0, 0 },
	{  13, "XEGS 64 KB cartridge (banks 0-7)",          FALSE, 0, 0, 0 },
	{  14, "XEGS 128 KB cartridge",                     FALSE, 0, 0, 0 },
	{  15, "OSS one chip 16 KB cartridge",              FALSE, 0, 0, 0 },
	{  16, "One chip 16 KB 5200 cartridge",             FALSE, 0, 0, 0 },
	{  17, "Decoded Atrax 128 KB cartridge",            FALSE, 0, 0, 0 },
	{  18, "Bounty Bob 40 KB cartridge",                FALSE, 0, 0, 0 },
	{  19, "Standard 8 KB 5200 cartridge",              FALSE, 0, 0, 0 },

	{  20, "Standard 4 KB 5200 cartridge",              FALSE, 0, 0, 0 },
	{  21, "Right slot 8 KB cartridge",                 FALSE, 0, 0, 0 },
	{  22, "32 KB Williams cartridge",                  FALSE, 0, 0, 0 },
	{  23, "XEGS 256 KB cartridge",                     FALSE, 0, 0, 0 },
	{  24, "XEGS 512 KB cartridge",                     FALSE, 0, 0, 0 },
	{  25, "XEGS 1 MB cartridge",                       FALSE, 0, 0, 0 },
	{  26, "MegaCart 16 KB cartridge",                  FALSE, 0, 0, 0 },
	{  27, "MegaCart 32 KB cartridge",                  FALSE, 0, 0, 0 },
	{  28, "MegaCart 64 KB cartridge",                  FALSE, 0, 0, 0 },
	{  29, "MegaCart 128 KB cartridge",                 FALSE, 0, 0, 0 },

	{  30, "MegaCart 256 KB cartridge",                 FALSE, 0, 0, 0 },
	{  31, "MegaCart 512 KB cartridge",                 FALSE, 0, 0, 0 },
	{  32, "MegaCart 1 MB cartridge",                   FALSE, 0, 0, 0 },
	{  33, "Switchable XEGS 32 KB cartridge",           FALSE, 0, 0, 0 },
	{  34, "Switchable XEGS 64 KB cartridge",           FALSE, 0, 0, 0 },
	{  35, "Switchable XEGS 128 KB cartridge",          FALSE, 0, 0, 0 },
	{  36, "Switchable XEGS 256 KB cartridge",          FALSE, 0, 0, 0 },
	{  37, "Switchable XEGS 512 KB cartridge",          FALSE, 0, 0, 0 },
	{  38, "Switchable XEGS 1 MB cartridge",            FALSE, 0, 0, 0 },
	{  39, "Phoenix 8 KB cartridge",                    FALSE, 0, 0, 0 },

	{  40, "Blizzard 16 KB cartridge",                  FALSE, 0, 0, 0 },
	{  41, "Atarimax 128 KB Flash cartridge",           FALSE, 0, 0, 0 },
	{  42, "Atarimax 1 MB Flash cartridge (old)",       FALSE, 0, 0, 0 },
	{  43, "SpartaDOS X 128 KB cartridge",              FALSE, 0, 0, 0 },
	{  44, "OSS 8 KB cartridge",                        FALSE, 0, 0, 0 },
	{  45, "OSS two chip 16 KB cartridge (043M)",       FALSE, 0, 0, 0 },
	{  46, "Blizzard 4 KB cartridge",                   FALSE, 0, 0, 0 },
	{  47, "AST 32 KB cartridge",                       FALSE, 0, 0, 0 },
	{  48, "Atrax SDX 64 KB cartridge",                 FALSE, 0, 0, 0 },
	{  49, "Atrax SDX 128 KB cartridge",                FALSE, 0, 0, 0 },

	{  50, "Turbosoft 64 KB cartridge",                 FALSE, 0, 0, 0 },
	{  51, "Turbosoft 128 KB cartridge",                FALSE, 0, 0, 0 },
	{  52, "Ultracart 32 KB cartridge",                 FALSE, 0, 0, 0 },
	{  53, "Low bank 8 KB cartridge",                   FALSE, 0, 0, 0 },
	{  54, "SIC! 128 KB cartridge",                     FALSE, 0, 0, 0 },
	{  55, "SIC! 256 KB cartridge",                     FALSE, 0, 0, 0 },
	{  56, "SIC! 512 KB cartridge",                     FALSE, 0, 0, 0 },
	{  57, "Standard 2 KB cartridge",                   FALSE, 0, 0, 0 },
	{  58, "Standard 4 KB cartridge",                   FALSE, 0, 0, 0 },
	{  59, "Right slot 4 KB cartridge",                 FALSE, 0, 0, 0 },

	{  60, "Blizzard 32 KB cartridge",                  FALSE, 0, 0, 0 },
	{  61, "MegaMax 2 MB cartridge",                    FALSE, 0, 0, 0 },
	{  62, "The!Cart 128 MB cartridge",                 FALSE, 0, 0, 0 },
	{  63, "Flash MegaCart 4 MB cartridge",             FALSE, 0, 0, 0 },
	{  64, "MegaCart 2 MB cartridge",                   FALSE, 0, 0, 0 },
	{  65, "The!Cart 32 MB cartridge",                  FALSE, 0, 0, 0 },
	{  66, "The!Cart 64 MB cartridge",                  FALSE, 0, 0, 0 },
	{  67, "XEGS 64 KB cartridge (banks 8-15)",         FALSE, 0, 0, 0 },
	{  68, "Atrax 128 KB cartridge",                    FALSE, 0, 0, 0 },
	{  69, "aDawliah 32 KB cartridge",                  FALSE, 0, 0, 0 },

	{  70, "aDawliah 64 KB cartridge",                  FALSE, 0, 0, 0 },
	{  71, "Super Cart 64 KB 5200 cartridge",           FALSE, 0, 0, 0 },
	{  72, "Super Cart 128 KB 5200 cartridge",          FALSE, 0, 0, 0 },
	{  73, "Super Cart 256 KB 5200 cartridge",          FALSE, 0, 0, 0 },
	{  74, "Super Cart 512 KB 5200 cartridge",          FALSE, 0, 0, 0 },
	{  75, "Atarimax 1 MB Flash cartridge (new)",       FALSE, 0, 0, 0 },
	{  76, "16 KB Williams cartridge",                  FALSE, 0, 0, 0 },
	{  77, "MIO diagnostics 8KB cartridge",             FALSE, 0, 0, 0 },
	{  78, "Telelink II cartridge",                     FALSE, 0, 0, 0 },
	{  79, "Pronto cartridge",                          FALSE, 0, 0, 0 },

	{  80, "JRC64 cartridge (linear)",                  FALSE, 0, 0, 0 },
	{  81, "MDDOS cartridge",                           FALSE, 0, 0, 0 },
	{  82, "COS32 cartridge",                           FALSE, 0, 0, 0 },
	{  83, "SIC+ 1024 KB cartridge",                    FALSE, 0, 0, 0 },
	{  84, "Corina 1M + 8K EEPROM cartridge",           FALSE, 0, 0, 0 },
	{  85, "Corina 512K + 512K SRAM + 8K EEPROM cartridge", FALSE, 0, 0, 0 },
	{  86, "XE Multicart (8KB) cartridge",              FALSE, 0, 0, 0 },
	{  87, "XE Multicart (16KB) cartridge",             FALSE, 0, 0, 0 },
	{  88, "XE Multicart (32KB) cartridge",             FALSE, 0, 0, 0 },
	{  89, "XE Multicart (64KB) cartridge",             FALSE, 0, 0, 0 },

	{  90, "XE Multicart (128KB) cartridge",            FALSE, 0, 0, 0 },
	{  91, "XE Multicart (256KB) cartridge",            FALSE, 0, 0, 0 },
	{  92, "XE Multicart (512KB) cartridge",            FALSE, 0, 0, 0 },
	{  93, "XE Multicart (1024KB) cartridge",           FALSE, 0, 0, 0 },
	{  94, "Ram-Cart 64 KB cartridge",                  FALSE, 0, 0, 0 },
	{  95, "Ram-Cart 128 KB cartridge",                 FALSE, 0, 0, 0 },
	{  96, "Double Ram-Cart 256 KB cartridge",          FALSE, 0, 0, 0 },
	{  97, "Ram-Cart 1 MB cartridge",                   FALSE, 0, 0, 0 },
	{  98, "Ram-Cart 2 MB cartridge",                   FALSE, 0, 0, 0 },
	{  99, "Ram-Cart 4 MB cartridge",                   FALSE, 0, 0, 0 },

	{ 100, "Ram-Cart 8 MB cartridge",                   FALSE, 0, 0, 0 },
	{ 101, "Ram-Cart 16 MB cartridge",                  FALSE, 0, 0, 0 },
	{ 102, "Ram-Cart 32 MB cartridge",                  FALSE, 0, 0, 0 },
	{ 103, "SiDiCar 32 KB cartridge",                   FALSE, 0, 0, 0 },
	{ 104, NULL,                                        FALSE, 0, 0, 0 },
	{ 105, NULL,                                        FALSE, 0, 0, 0 },
	{ 106, NULL,                                        FALSE, 0, 0, 0 },
	{ 107, NULL,                                        FALSE, 0, 0, 0 },
	{ 108, NULL,                                        FALSE, 0, 0, 0 },
	{ 109, NULL,                                        FALSE, 0, 0, 0 },

	{ 110, NULL,                                        FALSE, 0, 0, 0 },
	{ 111, NULL,                                        FALSE, 0, 0, 0 },
	{ 112, NULL,                                        FALSE, 0, 0, 0 },
	{ 113, NULL,                                        FALSE, 0, 0, 0 },
	{ 114, NULL,                                        FALSE, 0, 0, 0 },
	{ 115, NULL,                                        FALSE, 0, 0, 0 },
	{ 116, NULL,                                        FALSE, 0, 0, 0 },
	{ 117, NULL,                                        FALSE, 0, 0, 0 },
	{ 118, NULL,                                        FALSE, 0, 0, 0 },
	{ 119, NULL,                                        FALSE, 0, 0, 0 },

	{ 120, NULL,                                        FALSE, 0, 0, 0 },
	{ 121, NULL,                                        FALSE, 0, 0, 0 },
	{ 122, NULL,                                        FALSE, 0, 0, 0 },
	{ 123, NULL,                                        FALSE, 0, 0, 0 },
	{ 124, NULL,                                        FALSE, 0, 0, 0 },
	{ 125, NULL,                                        FALSE, 0, 0, 0 },
	{ 126, NULL,                                        FALSE, 0, 0, 0 },
	{ 127, NULL,                                        FALSE, 0, 0, 0 },
	{ 128, NULL,                                        FALSE, 0, 0, 0 },
	{ 129, NULL,                                        FALSE, 0, 0, 0 },

	{ 130, NULL,                                        FALSE, 0, 0, 0 },
	{ 131, NULL,                                        FALSE, 0, 0, 0 },
	{ 132, NULL,                                        FALSE, 0, 0, 0 },
	{ 133, NULL,                                        FALSE, 0, 0, 0 },
	{ 134, NULL,                                        FALSE, 0, 0, 0 },
	{ 135, NULL,                                        FALSE, 0, 0, 0 },
	{ 136, NULL,                                        FALSE, 0, 0, 0 },
	{ 137, NULL,                                        FALSE, 0, 0, 0 },
	{ 138, NULL,                                        FALSE, 0, 0, 0 },
	{ 139, NULL,                                        FALSE, 0, 0, 0 },

	{ 140, NULL,                                        FALSE, 0, 0, 0 },
	{ 141, NULL,                                        FALSE, 0, 0, 0 },
	{ 142, NULL,                                        FALSE, 0, 0, 0 },
	{ 143, NULL,                                        FALSE, 0, 0, 0 },
	{ 144, NULL,                                        FALSE, 0, 0, 0 },
	{ 145, NULL,                                        FALSE, 0, 0, 0 },
	{ 146, NULL,                                        FALSE, 0, 0, 0 },
	{ 147, NULL,                                        FALSE, 0, 0, 0 },
	{ 148, NULL,                                        FALSE, 0, 0, 0 },
	{ 149, NULL,                                        FALSE, 0, 0, 0 },

	{ 150, NULL,                                        FALSE, 0, 0, 0 },
	{ 151, NULL,                                        FALSE, 0, 0, 0 },
	{ 152, NULL,                                        FALSE, 0, 0, 0 },
	{ 153, NULL,                                        FALSE, 0, 0, 0 },
	{ 154, NULL,                                        FALSE, 0, 0, 0 },
	{ 155, NULL,                                        FALSE, 0, 0, 0 },
	{ 156, NULL,                                        FALSE, 0, 0, 0 },
	{ 157, NULL,                                        FALSE, 0, 0, 0 },
	{ 158, NULL,                                        FALSE, 0, 0, 0 },
	{ 159, NULL,                                        FALSE, 0, 0, 0 },

	{ 160, "JRC64 cartridge (interleaved)",             FALSE, 0, 0, 0 },
	{ 161, NULL,                                        FALSE, 0, 0, 0 },
	{ 162, NULL,                                        FALSE, 0, 0, 0 },
	{ 163, NULL,                                        FALSE, 0, 0, 0 },
	{ 164, NULL,                                        FALSE, 0, 0, 0 },
	{ 165, NULL,                                        FALSE, 0, 0, 0 },
	{ 166, NULL,                                        FALSE, 0, 0, 0 },
	{ 167, NULL,                                        FALSE, 0, 0, 0 },
	{ 168, NULL,                                        FALSE, 0, 0, 0 },
	{ 169, NULL,                                        FALSE, 0, 0, 0 },

	{ 170, NULL,                                        FALSE, 0, 0, 0 },
	{ 171, NULL,                                        FALSE, 0, 0, 0 },
	{ 172, NULL,                                        FALSE, 0, 0, 0 },
	{ 173, NULL,                                        FALSE, 0, 0, 0 },
	{ 174, NULL,                                        FALSE, 0, 0, 0 },
	{ 175, NULL,                                        FALSE, 0, 0, 0 },
	{ 176, NULL,                                        FALSE, 0, 0, 0 },
	{ 177, NULL,                                        FALSE, 0, 0, 0 },
	{ 178, NULL,                                        FALSE, 0, 0, 0 },
	{ 179, NULL,                                        FALSE, 0, 0, 0 },

	{ 180, NULL,                                        FALSE, 0, 0, 0 },
	{ 181, NULL,                                        FALSE, 0, 0, 0 },
	{ 182, NULL,                                        FALSE, 0, 0, 0 },
	{ 183, NULL,                                        FALSE, 0, 0, 0 },
	{ 184, NULL,                                        FALSE, 0, 0, 0 },
	{ 185, NULL,                                        FALSE, 0, 0, 0 },
	{ 186, NULL,                                        FALSE, 0, 0, 0 },
	{ 187, NULL,                                        FALSE, 0, 0, 0 },
	{ 188, NULL,                                        FALSE, 0, 0, 0 },
	{ 189, NULL,                                        FALSE, 0, 0, 0 },

	{ 190, NULL,                                        FALSE, 0, 0, 0 },
	{ 191, NULL,                                        FALSE, 0, 0, 0 },
	{ 192, NULL,                                        FALSE, 0, 0, 0 },
	{ 193, NULL,                                        FALSE, 0, 0, 0 },
	{ 194, NULL,                                        FALSE, 0, 0, 0 },
	{ 195, NULL,                                        FALSE, 0, 0, 0 },
	{ 196, NULL,                                        FALSE, 0, 0, 0 },
	{ 197, NULL,                                        FALSE, 0, 0, 0 },
	{ 198, NULL,                                        FALSE, 0, 0, 0 },
	{ 199, NULL,                                        FALSE, 0, 0, 0 },

	{ 200, NULL,                                        FALSE, 0, 0, 0 },
	{ 201, NULL,                                        FALSE, 0, 0, 0 },
	{ 202, NULL,                                        FALSE, 0, 0, 0 },
	{ 203, NULL,                                        FALSE, 0, 0, 0 },
	{ 204, NULL,                                        FALSE, 0, 0, 0 },
	{ 205, NULL,                                        FALSE, 0, 0, 0 },
	{ 206, NULL,                                        FALSE, 0, 0, 0 },
	{ 207, NULL,                                        FALSE, 0, 0, 0 },
	{ 208, NULL,                                        FALSE, 0, 0, 0 },
	{ 209, NULL,                                        FALSE, 0, 0, 0 },

	{ 210, NULL,                                        FALSE, 0, 0, 0 },
	{ 211, NULL,                                        FALSE, 0, 0, 0 },
	{ 212, NULL,                                        FALSE, 0, 0, 0 },
	{ 213, NULL,                                        FALSE, 0, 0, 0 },
	{ 214, NULL,                                        FALSE, 0, 0, 0 },
	{ 215, NULL,                                        FALSE, 0, 0, 0 },
	{ 216, NULL,                                        FALSE, 0, 0, 0 },
	{ 217, NULL,                                        FALSE, 0, 0, 0 },
	{ 218, NULL,                                        FALSE, 0, 0, 0 },
	{ 219, NULL,                                        FALSE, 0, 0, 0 },

	{ 220, NULL,                                        FALSE, 0, 0, 0 },
	{ 221, NULL,                                        FALSE, 0, 0, 0 },
	{ 222, NULL,                                        FALSE, 0, 0, 0 },
	{ 223, NULL,                                        FALSE, 0, 0, 0 },
	{ 224, NULL,                                        FALSE, 0, 0, 0 },
	{ 225, NULL,                                        FALSE, 0, 0, 0 },
	{ 226, NULL,                                        FALSE, 0, 0, 0 },
	{ 227, NULL,                                        FALSE, 0, 0, 0 },
	{ 228, NULL,                                        FALSE, 0, 0, 0 },
	{ 229, NULL,                                        FALSE, 0, 0, 0 },

	{ 230, NULL,                                        FALSE, 0, 0, 0 },
	{ 231, NULL,                                        FALSE, 0, 0, 0 },
	{ 232, NULL,                                        FALSE, 0, 0, 0 },
	{ 233, NULL,                                        FALSE, 0, 0, 0 },
	{ 234, NULL,                                        FALSE, 0, 0, 0 },
	{ 235, NULL,                                        FALSE, 0, 0, 0 },
	{ 236, NULL,                                        FALSE, 0, 0, 0 },
	{ 237, NULL,                                        FALSE, 0, 0, 0 },
	{ 238, NULL,                                        FALSE, 0, 0, 0 },
	{ 239, NULL,                                        FALSE, 0, 0, 0 },

	{ 240, NULL,                                        FALSE, 0, 0, 0 },
	{ 241, NULL,                                        FALSE, 0, 0, 0 },
	{ 242, NULL,                                        FALSE, 0, 0, 0 },
	{ 243, NULL,                                        FALSE, 0, 0, 0 },
	{ 244, NULL,                                        FALSE, 0, 0, 0 },
	{ 245, NULL,                                        FALSE, 0, 0, 0 },
	{ 246, NULL,                                        FALSE, 0, 0, 0 },
	{ 247, NULL,                                        FALSE, 0, 0, 0 },
	{ 248, NULL,                                        FALSE, 0, 0, 0 },
	{ 249, NULL,                                        FALSE, 0, 0, 0 },

	{ 250, NULL,                                        FALSE, 0, 0, 0 },
	{ 251, NULL,                                        FALSE, 0, 0, 0 },
	{ 252, NULL,                                        FALSE, 0, 0, 0 },
	{ 253, NULL,                                        FALSE, 0, 0, 0 },
	{ 254, NULL,                                        FALSE, 0, 0, 0 },
	{ 255, "MaxPlus 512K cartridge",                    TRUE,  1, 0x1000, 0x80 },
};

/* Cartridge memory buffer. */
static char buffer[ BANK_SIZE * BANK_COUNT ];


vrhi_status_t device_reset(void)
{
	vrhi_status_t status;

	printf("Reset device\n");
	status = vrhi_reset();
	if (status != EOK)
	{
		return status;
	}

	return EOK;
}

vrhi_status_t device_info(void)
{
	vrhi_status_t status;
	vrhi_response_info_t response_info;

	printf("Reading device information\n");
	status = vrhi_info(&response_info);
	if (status != EOK)
	{
		return status;
	}
	printf("Signature: %.4s\n", response_info.signature);
	printf("Platform: %.2s\n", response_info.platform);
	printf("Firmware: %d.%d\n", response_info.version, response_info.revision);

	return EOK;
}

vrhi_status_t cartridge_power_on(void)
{
	vrhi_status_t status;
	vrhi_request_power_t request_power = {
		._vpp = LOW,	/* power on */
	};
	vrhi_response_power_t response_power;

	printf("Cartridge power on\n");

	status = vrhi_power(&request_power, &response_power);
	if (status != EOK)
	{
		return status;
	}
	if (response_power._vpp != LOW)
	{
		return EERRNO;	/*TODO: error code*/
	}
	if (usleep(20) != 0)
	{
		return EERRNO;
	}

	return status;
}

vrhi_status_t cartridge_power_off(void)
{
	vrhi_status_t status;
	vrhi_request_power_t request_power = {
		._vpp = HIGH,	/* power off */
	};
	vrhi_response_power_t response_power;

	printf("Cartridge power off\n");

	status = vrhi_power(&request_power, &response_power);
	if (status != EOK)
	{
		return status;
	}
	if (response_power._vpp != HIGH)
	{
		return EERRNO;	/*TODO: error code*/
	}
	if (usleep(20) != 0)
	{
		return EERRNO;
	}

	return status;
}

vrhi_status_t read_cartridge_memory_chip_identifier(const cartridge_t *cartridge)
{
	vrhi_status_t status;
	vrhi_request_chip_info_t request_info;
	vrhi_response_chip_info_t response_info;
	char *manufacturer_name, *device_name;
	unsigned int chip;

	for (chip = 0; chip < cartridge->chip_count; chip++)
	{
		printf("Reading cartridge memory chip %d identifier\n", 1+chip);

		request_info.address = chip * cartridge->sector_size * cartridge->sector_count;
		status = vrhi_chip_info(&request_info, &response_info);
		if (status != EOK)
		{
			return status;
		}

		switch (response_info.manufacturer_id)
		{
		case 0xad:
			manufacturer_name = "Hyundai Electronics";
			break;
		case 0xbf:
			manufacturer_name = "SST";
			break;
		default:
			manufacturer_name = "unknown";
		}

		switch (response_info.device_id)
		{
		case 0xa4:
			device_name = "HY29F040A";
			break;
		case 0xb5:
			device_name = "SST39SF010A";
			break;
		case 0xb6:
			device_name = "SST39SF020A";
			break;
		case 0xb7:
			device_name = "SST39SF040";
			break;
		default:
			device_name = "unknown";
		}

		printf("Manufacturer: %s ($%02x)\n", manufacturer_name, response_info.manufacturer_id);
		printf("Device: %s ($%02x)\n", device_name, response_info.device_id);
	}

	return EOK;
}

vrhi_status_t read_cartridge_memory_content_in_loop(unsigned int bank)
{
	vrhi_status_t status;
	vrhi_request_bank_read_t request_read;
	vrhi_response_bank_read_t response_read;
	unsigned int retry;

	for (retry = 0; retry < RETRY_COUNT; retry++)
	{
		printf(".");
		request_read.bank = bank;
		response_read.data = buffer + BANK_SIZE * bank;
		status = vrhi_bank_read(&request_read, &response_read);
		if (status == EOK)
		{
			return EOK;
		}
	}
	return status;
}

vrhi_status_t read_cartridge_memory_content(const cartridge_t *cartridge)
{
	vrhi_status_t status;
	unsigned char bank, bank_count;
	unsigned int cartridge_size;

	printf("Reading content of the cartridge\n");
	cartridge_size = cartridge->chip_count * cartridge->sector_count * cartridge->sector_size;
	bank_count = cartridge_size / BANK_SIZE;
	for (bank = 0; bank < bank_count; bank++)
	{
		printf("Reading memory bank $%02x/$%02x content\t", bank, bank_count);
		status = read_cartridge_memory_content_in_loop(bank);
		if (status != EOK)
		{
			printf("FAIL\n");
			break;
		}
		else
		{
			printf("OK\n");
		}
	}

	return status;
}

vrhi_status_t erase_cartridge_memory_content_in_loop(unsigned int chip, const cartridge_t *cartridge)
{
	vrhi_status_t status;
	vrhi_request_chip_erase_t request_erase;
	unsigned int retry;

	for (retry = 0; retry < RETRY_COUNT; retry++)
	{
		printf(".");
		request_erase.address = chip * cartridge->sector_count * cartridge->sector_size;
		status = vrhi_chip_erase(&request_erase);
		if (status == EOK)
		{
			return EOK;
		}
	}

	return status;
}

vrhi_status_t write_cartridge_memory_content_in_loop(unsigned int bank)
{
	vrhi_status_t status;
	vrhi_request_bank_write_t request_write;
	unsigned int retry;

	for (retry = 0; retry < RETRY_COUNT; retry++)
	{
		printf(".");
		request_write.bank = bank;
		request_write.data = buffer + BANK_SIZE * bank;
		status = vrhi_bank_write(&request_write);
		if (status == EOK)
		{
			return EOK;
		}
	}

	return status;
}

vrhi_status_t write_cartridge_memory_content(const cartridge_t *cartridge)
{
	vrhi_status_t status;
	unsigned char chip, bank, bank_count;
	unsigned int cartridge_size;

	for (chip = 0; chip < cartridge->chip_count; chip++)
	{
		printf("Erase chip %d on the cartridge\t", 1+chip);
		status = erase_cartridge_memory_content_in_loop(chip, cartridge);
		if (status != EOK)
		{
			printf("FAIL\n");
			return status;
		}
		else
		{
			printf("OK\n");
		}
	}

	printf("Writing content to the cartridge\n");
	cartridge_size = cartridge->chip_count * cartridge->sector_count * cartridge->sector_size;
	bank_count = cartridge_size / BANK_SIZE;
	for (bank = 0; bank < bank_count; bank++)
	{
		printf("Writing memory bank $%02x/$%02x content\t", bank, bank_count);
		status = write_cartridge_memory_content_in_loop(bank);
		if (status != EOK)
		{
			printf("FAIL\n");
			return status;
		}
		else
		{
			printf("OK\n");
		}
	}

	return status;
}

vrhi_status_t read_rom_file(char *filename)
{
	vrhi_status_t status;
	FILE *f;

	printf("Reading file %s\n", filename);
	f = fopen(filename, "rb");
	if (f != NULL)
	{
		ssize_t count, size;

		size = sizeof(buffer);

		memset(buffer, 0xff, size);

		count = fread(buffer, 1, size, f);
		if ((count & 0x1fff) != 0)
		{
			printf("File has too few bytes.\n");
			status = EOK;
		}
		else
		{
			status = EOK;
		}
		if (fclose(f) != 0)
		{
			status = EERRNO;
		}
	}
	else
	{
		status = EERRNO;
	}

	return status;
}

vrhi_status_t write_rom_file(char *filename)
{
	vrhi_status_t status;
	FILE *f;

	printf("Writing file %s\n", filename);
	f = fopen(filename, "wb");
	if (f != NULL)
	{
		ssize_t count, size;

		size = sizeof(buffer);
		count = fwrite(buffer, 1, size, f);
		if (count != size)
		{
			status = ESTOOFEW;
		}
		else
		{
			status = EOK;
		}
		if (fclose(f) != 0)
		{
			status = EERRNO;
		}
	}
	else
	{
		status = EERRNO;
	}

	return status;
}

vrhi_status_t process(char *operation, char *filename, const cartridge_t *cartridge)
{
	vrhi_status_t status;

	status = read_cartridge_memory_chip_identifier(cartridge);
	if (status != EOK)
	{
		return status;
	}

	switch (*operation)
	{
	case 'r'/*ead*/:
		status = read_cartridge_memory_content(cartridge);
		if (status != EOK)
		{
			return status;
		}
		status = write_rom_file(filename);
		break;
	case 'p'/*rogram*/:
		status = read_rom_file(filename);
		if (status != EOK)
		{
			return status;
		}
		status = write_cartridge_memory_content(cartridge);
		break;
	default:
		printf("Unexpected operation: %s\n", operation);
	}

	return status;
}

int main(int argc, char *argv[])
{
	vrhi_status_t status;
	const cartridge_t *cartridge;
	int opt, help, type, index;
	char *operation, *serial, *filename;

	printf("Virtual to Real Hardware Interface Cartridge Programator\n"
		"Library: %d.%d\n",
		vrhi_version, vrhi_revision);

	operation = "read";
	serial = DEFAULT_DEVICE;
	filename = NULL;
	help = FALSE;
	type = CART_NONE;
	while ((opt = getopt(argc, argv, "ps:t:")) >= 0)
	{
		switch (opt)
		{
		case 'p':
			operation = "program";
			break;
		case 's':
			serial = optarg;
			break;
		case 't':
			type = atoi(optarg);
			break;
		default:
			help = TRUE;
		}
	}
	if (optind < argc)
	{
		filename = argv[optind];
	}

	if ( help
		|| (filename == NULL)
	)
	{
		printf(
			"\n"
			"Usage: %s [-p] [-s device] [-t type] image.rom\n"
			"\n"
			"Where:\n"
			"-p\t\tprogram image.rom on the device\n"
			"\t\t(reading from the device into file is the default action)\n"
			"-s device\tthe serial device where programator device is connected to\n"
			"\t\t(" DEFAULT_DEVICE " is taken by default)\n"
			"-t type\t\tthe type of cartridge\n"
			"image.rom\tthe ROM image file\n"
			"\n"
			"Cartridge types supported:\n",
			argv[0]
		);
		for (index = 0; index < CART_COUNT; index++)
		{
			cartridge = &CARTRIDGES[index];
			if (cartridge->programmable)
			{
				printf("- %3d\t%s\n", index, cartridge->name);
			}
		}
		printf(
			"\n"
			"Send bug reports and feature requests to Jerzy Kut <mono@protolabo.pl>\n"
		);
		exit(EXIT_FAILURE);
	}

	if ((type > CART_NONE) && (type < CART_COUNT))
	{
		cartridge = &CARTRIDGES[type];

		printf("Cartridge: %d %s", cartridge->type, cartridge->name);
		if (cartridge->programmable)
		{
			printf("\n");

			printf("Opening %s\n", serial);
			status = vrhi_open(serial);
			if (status == EOK)
			{
				status = device_reset();
				if (status == EOK)
				{
					status = device_info();
					if (status == EOK)
					{
						vrhi_status_t status_off, status_reset;
			
						status = cartridge_power_on();
						if (status == EOK)
						{
							status = process(operation, filename, cartridge);
						}
						status_off = cartridge_power_off();
						if (status == EOK)
						{
							status = status_off;
						}
						status_reset = device_reset();
						if (status == EOK)
						{
							status = status_reset;
						}
					}
				}
			}
			if (status != EOK)
			{
				char *errstr;
	
				switch (status)
				{
				case EWPLATFORM:
					errstr = "Wrong device platform";
					break;
				case EWSIGNATURE:
					errstr = "Wrong device signature";
					break;
				case EWCSUM:
					errstr = "Wrong checksum";
					break;
				case ENEXISTS:
					errstr = "Device not exists";
					break;
				case ERTOOFEW:
					errstr = "Too few data received";
					break;
				case ESTOOFEW:
					errstr = "Too few data sent";
					break;
				case ENOPENED:
					errstr = "Device not opened";
					break;
				case EAOPENED:
					errstr = "Device already opened";
					break;
				case EERRNO:
					errstr = strerror(errno);
					status = errno;
					break;
				default:
					errstr = "Unknown error";
				}
				printf("Error %d: %s\n", status, errstr);
			}

			printf("Closing %s\n", serial);
			vrhi_close();
		}
		else
		{
			printf(" is not programmable\n");
		}
	}

	exit(EXIT_SUCCESS);
}

